# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2017 McAfee Inc. - All Rights Reserved.
################################################################################
import json
import logging
import weakref
from threading import Timer, Condition, RLock
import time

from dxlclient import _BaseObject
from dxlclient._callback_manager import _RequestCallbackManager
from dxlclient._uuid_generator import UuidGenerator
from dxlclient.callbacks import RequestCallback
from dxlclient.exceptions import DxlException
from dxlclient.message import Message, Request, ErrorResponse

logger = logging.getLogger(__name__)


class ServiceRegistrationInfo(_BaseObject):
    """
    Service Registration instances are used to register and expose services onto a DXL fabric.

    DXL Services are exposed to the DXL fabric and are invoked in a fashion similar to RESTful web services.
    Communication between an invoking client and the DXL service is one-to-one (request/response).

    Each service is identified by the "topics" it responds to. Each of these "topics" can be thought of as
    a method that is being "invoked" on the service by the remote client.

    Multiple service "instances" can be registered with the DXL fabric that respond to the same "topics". When
    this occurs (unless explicitly overridden by the client) the fabric will select the particular instance
    to route the request to (by default round-robin). Multiple service instances can be used to increase
    scalability and fault-tolerance.

    The following demonstrates registering a service that responds to a single topic with the DXL fabric:

    .. code-block:: python

        from dxlclient.callbacks import RequestCallback
        from dxlclient.message import Response
        from dxlclient.service import ServiceRegistrationInfo

        class MyRequestCallback(RequestCallback):
            def on_request(self, request):
                # Extract information from request
                print request.payload.decode()

                # Create the response message
                res = Response(request)

                # Populate the response payload
                res.payload = "pong".encode()

                # Send the response
                dxl_client.send_response(res)

        # Create service registration object
        info = ServiceRegistrationInfo(dxl_client, "/mycompany/myservice")

        # Add a topic for the service to respond to
        info.add_topic("/testservice/testrequesttopic", MyRequestCallback())

        # Register the service with the fabric (wait up to 10 seconds for registration to complete)
        dxl_client.register_service_sync(info, 10)

    **NOTE:** A service is only considered "active" if there are references to the
    :class:`ServiceRegistrationInfo` instance. If no references to the info object exist, it will be
    destructed, and the service will be automatically *unregistered* from the fabric.

    The following demonstrates a client that is invoking the service in the example above:

    .. code-block:: python

        from dxlclient.message import Request, Message

        # Create the request message
        req = Request("/testservice/testrequesttopic")

        # Populate the request payload
        req.payload = "ping".encode()

        # Send the request and wait for a response (synchronous)
        res = dxl_client.sync_request(req)

        # Extract information from the response (if an error did not occur)
        if res.message_type != Message.MESSAGE_TYPE_ERROR:
            print res.payload.decode()
    """

    def __init__(self, client, service_type):
        """
        Constructor parameters:
        
        :param client: The :class:`dxlclient.client.DxlClient` instance that will expose this service
        :param service_type: A textual name for the service. For example, "/mycompany/myservice"
        """
        super(ServiceRegistrationInfo, self).__init__()

        # if not isinstance(channels, list):
        # raise ValueError('Channels should be a list')
        # if not channels:
        #     raise InvalidServiceException('Channel list is empty')
        if not service_type:
            raise ValueError("Undefined service name")

            #The service type or name prefix
        self._service_type = service_type
        # The unique service ID
        self._service_id = UuidGenerator.generate_id_as_string()

        # The map of registered channels and their associated callbacks
        self._callbacks_by_topic = {}
        #The map of meta data associated with this service (name-value pairs)
        self._metadata = {}

        # The Time-To-Live (TTL) of the service registration (default: 60 minutes)
        self._ttl = 60  # minutes
        # The minimum Time-To-Live (TTL) of the service registration (default: 10 minutes)
        self._ttl_lower_limit = 10

        # Internal client reference
        self._dxl_client = client

        # Registration sync object
        self._registration_sync = Condition()

        # Whether at least one registration has occurred
        self._registration_occurred = False

        # Whether at least one unregistration registration has occurred
        self._unregistration_occurred = False

        self._destroy_lock = RLock()
        self._destroyed = False

    def __del__(self):
        """destructor"""
        super(ServiceRegistrationInfo, self).__del__()
        self._destroy()

    def _destroy(self, unregister=True):
        """
        Destroys the service registration

        :param unregister: Whether to unregister the service from the fabric
        """
        with self._destroy_lock:
            if not self._destroyed:
                if unregister and self._dxl_client:
                    try:
                        self._dxl_client.unregister_service_async(self)
                    except Exception:
                        # Currently ignoring this as it can occur due to the fact that we are
                        # attempting to unregister a service that was never registered
                        pass
                self._dxl_client = None
                self._destroyed = True

    @property
    def service_type(self):
        """
        A textual name for the service. For example, "/mycompany/myservice"
        """
        return self._service_type

    @property
    def service_id(self):
        """
        A unique identifier for the service instance (automatically generated when the :class:`ServiceRegistrationInfo`
        object is constructed)
        """
        return self._service_id

    @property
    def metadata(self):
        """
        A dictionary of name-value pairs that are sent as part of the service registration. Brokers provide
        a registry service that allows for registered services and their associated meta-information to be inspected.
        The metadata is typically used to include information such as the versions for products that are
        exposing DXL services, etc.
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = metadata

    @property
    def ttl(self):
        """
        The interval (in minutes) at which the client will automatically re-register the service with the
        DXL fabric (defaults to 60 minutes)
        """
        return self._ttl

    @ttl.setter
    def ttl(self, ttl):
        self._ttl = ttl

    @property
    def topics(self):
        """
        Returns a tuple containing the topics that the service responds to
        """
        return tuple(self._callbacks_by_topic.keys())

    def add_topic(self, topic, callback):
        """
        Registers a topic for the service to respond to along with the :class:`dxlclient.callbacks.RequestCallback`
        that will be invoked.
        
        :param topic:  The topic for the service to respond to
        :param callback: The :class:`dxlclient.callbacks.RequestCallback` that will be invoked when a
            :class:`dxlclient.message.Request` message is received
        """
        # TODO: use dictionary get method
        try:
            callbacks = self._callbacks_by_topic[topic]
        except KeyError:
            callbacks = set()
            self._callbacks_by_topic[topic] = callbacks
        finally:
            callbacks.add(callback)

    def add_topics(self, callbacks_by_topic):
        """
        Registers a set of topics for the service to respond to along with their associated
        :class:`dxlclient.callbacks.RequestCallback` instances as a dictionary

        :param callbacks_by_topic: Dictionary containing a set of topics for the service to respond to along with
            their associated :class:`dxlclient.callbacks.RequestCallback` instances
        """
        if not isinstance(callbacks_by_topic, dict):
            raise ValueError("Channel and callback should be a dictionary")
        if not callbacks_by_topic:
            raise ValueError("Undefined channel")
        for channel, callback in callbacks_by_topic.iteritems():
            self.add_topic(channel, callback)

    def _wait_for_registration_notification(self, wait_time, is_register):
        """
        Waits for a registration notification (register or unregister).
        
        :param waitTime:   The amount of time to wait.
        :param isRegister Whether we are waiting for a register or unregister notification.
        :return: None.
        """
        with self._registration_sync:
            if wait_time > 0:
                self._registration_sync.wait(wait_time)
            else:
                raise DxlException("Timeout waiting for service related notification")

    def _wait_for_registration(self, timeout):
        """
        Waits for the service to be registered with the broker for the first time.
        
        :param timeout: The amount of time to wait for the registration to occur.
        :return: None.
        """
        with self._registration_sync:
            end_time = int(time.time()) + timeout
            while not self._registration_occurred:
                self._wait_for_registration_notification(end_time - int(time.time()), True)

    def _notify_registration_succeeded(self):
        """
        Invoked when the service has been successfully registered with a broker.
        
        :return: None.
        """
        with self._registration_sync:
            self._registration_occurred = True
            self._unregistration_occurred = False
            self._registration_sync.notify_all()

    # @synchronized
    def _wait_for_unregistration(self, timeout):
        """
        Waits for the service to be unregistered with the broker for the first time.
         
        :param timeout: The amount of time to wait for the unregistration to occur.
        :return: None.
        """
        end_time = int(time.time()) + timeout
        with self._registration_sync:
            while not self._unregistration_occurred:
                self._wait_for_registration_notification(end_time - int(time.time()), False)

    #@synchronized
    def _notify_unregistration_succeeded(self):
        """
        Invoked when the service has been successfully unregistered with a broker.
         
        :return: None.
        """
        with self._registration_sync:
            self._registration_occurred = False
            self._unregistration_occurred = True
            self._registration_sync.notify_all()


class _ServiceRegistrationHandler(_BaseObject):
    def __init__(self, client, service):
        """
        Constructs the ServiceRegistrationHandler object.
        
        :param client:  The internal client reference.
        :param service: The service registration info.
        :return: None.
        """
        super(_ServiceRegistrationHandler, self).__init__()

        # The service type or name prefix
        self.service_type = service.service_type
        # The service instance ID
        self.instance_id = service.service_id
        # The list of full qualified registered channels
        self.channels = service.topics
        # The map of meta data associated with this service (name-value pairs)
        self.metadata = service.metadata.copy()
        # The Time-To-Live (TTL) grace period of the service registration (default: 10 minutes)
        self.ttl_grace_period = 10
        # The service registration info (weak reference) */
        self.service_weak_reference = weakref.ref(service)
        # The internal client reference */
        self.client = client
        # The request callback manager */
        self.request_callbacks = _RequestCallbackManager()
        self.deleted = False
        self.ttl_timer = None
        self.register_time = 0
        self.ttl = service.ttl
        self.lock = RLock()

        self._destroy_lock = RLock()
        self._destroyed = False

        service_ref = self.get_service()
        if not service_ref:
            raise DxlException("Service no longer valid")

        for channel, callbacks in service_ref._callbacks_by_topic.iteritems():
            for callback in callbacks:
                self.request_callbacks.add_callback(channel, callback)

    def __del__(self):
        """destructor"""
        super(_ServiceRegistrationHandler, self).__del__()
        self.destroy()

    def destroy(self, unregister=True):
        """
        Destroys the service registration

        :param unregister: Whether to unregister the service from the fabric
        """
        with self._destroy_lock:
            if not self._destroyed:
                info = self.service_weak_reference()
                if info:
                    for channel, callbacks in info._callbacks_by_topic.iteritems():
                        for callback in callbacks:
                            self.request_callbacks.remove_callback(channel, callback)
                    info._destroy(unregister)
                self.client = None
                self._destroyed = True

    def get_service(self):
        """

        Returns the service registration info.
        
        :return The service registration info.
        """
        return self.service_weak_reference()

    def is_invalid_reference(self):
        """
        Returns true if the service weak reference is invalid.
        
        :return True if the service weak reference is invalid, otherwise false.
        """
        return self.get_service() and None

    def send_register_service_request(self):
        """
        Send the registration request for the service.
        
        :return: None.
        """
        if not self.client:
            raise DxlException("Client not defined")
        req = Request(destination_topic=_ServiceManager.DXL_SERVICE_REGISTER_REQUEST_CHANNEL)
        req.payload = bytes(self.json_register_service())
        response = self.client.sync_request(req, timeout=10)

        if response.message_type != Message.MESSAGE_TYPE_ERROR:
            self.update_register_time()
            info = self.get_service()
            if info:
                info._notify_registration_succeeded()
        else:
            # TODO: Notify the client with an exception if an error occurred, so that it doesn't wait for timeout
            logger.error("Error registering service.")

    def send_unregister_service_event(self):
        """
        Send the unregister event for the service.
        
        :return: None.
        """
        if not self.client:
            raise DxlException("Client not defined")
        # Send the unregister event only if the register event was sent before and TTL has not yet expired.
        current_time = int(time.time())
        last_register_time = self.get_register_time()

        if last_register_time > 0 and (current_time - last_register_time) <= (self.ttl * 60):
            request = Request(destination_topic=_ServiceManager.DXL_SERVICE_UNREGISTER_REQUEST_CHANNEL)
            request.payload = bytes(self.json_unregister_service())
            response = self.client.sync_request(request, timeout=60)
            if response.message_type != Message.MESSAGE_TYPE_ERROR:
                info = self.get_service()
                if info:
                    info._notify_unregistration_succeeded()
            else:
                raise DxlException("Unregister service request timed out")
        else:
            raise DxlException("Unregister service request timed out")

    def get_register_time(self):
        """
        Returns the last registration time in milliseconds.
        
        :return The last registration time in milliseconds, or 0L if not registered.
        """
        with self.lock:
            return self.register_time

    def update_register_time(self):
        """
        Updates the last registration time in milliseconds.
        
        :return: None.
        """
        with self.lock:
            self.register_time = int(time.time())

    def mark_for_deletion(self):
        """
        Marks the service for deletion.
        
        :return: None.
        """
        self.deleted = True

    def is_deleted(self):
        """
        Returns true if the service is marked for deletion.
        
        :return True if the service is marked for deletion, otherwise false.
        """
        return self.deleted

    def _timer_callback(self):
        """Callback invoked by the timer task (to re-register the service"""
        if self.client.connected:
            # Send unregister event if service marked for deletion or is no longer valid
            if self.deleted or self.is_invalid_reference():
                self.mark_for_deletion()
                self.send_unregister_service_event()
                self.ttl_timer.cancel()
            else:
                self.send_register_service_request()
                self.ttl_timer = Timer(self.ttl * 60, self._timer_callback)
                self.ttl_timer.daemon = True
                self.ttl_timer.start()
        else:
            if self.ttl_timer:
                self.ttl_timer.cancel()

    def start_timer(self):
        """
        Starts the TTL timer task.
        
        :return: None.
        """
        if not self.client:
            raise DxlException("Client not defined")

        if self.client.connected and not self.deleted:
            self.ttl_timer = Timer(0, self._timer_callback)
            self.ttl_timer.daemon = True
            self.ttl_timer.start()

    def stop_timer(self):
        """
        Stops the TTL timer task.
        
        :return: None.
        """
        if self.ttl_timer:
            self.ttl_timer.cancel()
            self.ttl_timer = None

    def json_register_service(self):  # instanceId or service guid?
        return json.dumps({
            'serviceType': self.service_type,
            'metaData': self.metadata,
            'requestChannels': list(self.channels),
            'ttlMins': self.ttl,
            'serviceGuid': self.instance_id
        })

    def json_unregister_service(self):
        return json.dumps({
            'serviceGuid': self.instance_id
        })


# pylint: disable= invalid-name
class _ServiceManager(RequestCallback):
    DXL_SERVICE_UNREGISTER_REQUEST_CHANNEL = "/mcafee/service/dxl/svcregistry/unregister"
    # The channel notified when services are registered
    DXL_SERVICE_REGISTER_CHANNEL = "/mcafee/event/dxl/svcregistry/register"
    # The channel to publish a service registration request to
    DXL_SERVICE_REGISTER_REQUEST_CHANNEL = "/mcafee/service/dxl/svcregistry/register"
    # The channel notified when services are unregistered
    DXL_SERVICE_UNREGISTER_CHANNEL = "/mcafee/event/dxl/svcregistry/unregister"

    def __init__(self, client):
        """
        Creates the service manager.
        
        :param client: client that the service manager is associated with.
        :return: None.
        """
        super(_ServiceManager, self).__init__()

        # if not isinstance(client, DxlClient):
        #     raise ValueError("Expected DxlClient object")
        self.__client = client
        #Map containing registered services */
        self.services = {}

        self.lock = RLock()

    def destroy(self):
        """Destroys the service manager (releases resources)"""
        with self.lock:
            for key in self.services.keys():
                self.remove_service(key)
            self.__client = None

    def add_service(self, service_reg_info):
        """
        Adds the specified service.
        
        :param service: The service to add.
        :return: None.
        """
        if not isinstance(service_reg_info, (type, ServiceRegistrationInfo)):
            raise ValueError("Expected ServiceRegistrationInfo object")

        with self.lock:
            try:
                service_handler = self.services[service_reg_info._service_id]
                raise DxlException("Service already registered")
            except KeyError:
                pass

            # Add service to registry
            service_handler = _ServiceRegistrationHandler(self.__client, service_reg_info)
            self.services[service_reg_info._service_id] = service_handler

            # Subscribe channels
            for channel in service_reg_info.topics:
                self.__client.subscribe(channel)
                self.__client.add_request_callback(channel, self)

            if self.__client.connected:
                service_handler.start_timer()

    def remove_service(self, service_id):
        """
        Removes the specified service.
        
        :param instanceId: The instance ID of the service to remove.
        :return: None.
        """
        if not isinstance(service_id, (type, basestring)):
            raise ValueError("Expected service id")

        if not service_id:
            raise ValueError("Invalid service id")

        with self.lock:
            service_handler = self.services.get(service_id, None)
            if not service_handler:
                raise DxlException("Service instance ID unknown: " + str(service_id))

            service_handler.stop_timer()

            for channel in service_handler.channels:
                self.__client.unsubscribe(channel)
                self.__client.remove_request_callback(channel, self)

            service_handler.mark_for_deletion()

            #If the client is actually connected, send unregister event. Remove upon success.
            if self.__client.connected:
                try:
                    service_handler.send_unregister_service_event()
                except Exception, ex:
                    logger.error("Error sending unregister service event for " +
                         service_handler.service_type + " (" + service_handler.instance_id + "): " + str(ex))

            del self.services[service_id]
            service_handler.destroy(unregister=False)

    def on_request(self, request):
        """
        Invoked when a {@link Request} has been received.
        
        :param request: The request.
        :return: None.
        """
        with self.lock:
            service_instance_id = request.service_id
            if not service_instance_id:
                for service in self.services:
                    self._on_request(self.services[service], request)
            else:
                service_registration_handler = self.services.get(service_instance_id)
                if service_registration_handler:
                    self._on_request(service_registration_handler, request)
                else:
                    logger.warning("No service with GUID " + service_instance_id + " registered. Ignoring request.")
                    self.send_service_not_found_error_message(request)

    def send_service_not_found_error_message(self, request):
        """
        Sends a service not found error message response.
        
        :param request: The request.
        :return: None.
        """
        # todo: error code via constants
        error_response = ErrorResponse(request=request, error_code=0x80000001,
                                       error_message="unable to locate service for request")

        try:
            self.__client.send_response(error_response)
        except Exception, ex:
            logger.error("Error sending service not found error message: " + str(ex))

    def _on_request(self, service_handler, request):
        service_handler.request_callbacks.fire_message(request)

    def on_connect(self):
        """
        On connect, check for deleted and/or rogue services and send unregister event, if necessary.
        Start timer threads for all active services.
        
        :return: None
        """
        with self.lock:
            for service_id in self.services:
                if self.services[service_id].is_deleted() or self.services[service_id].is_invalid_reference():
                    try:
                        self.services[service_id].send_unregister_service_event()
                        del self.services[service_id]
                    except Exception, ex:
                        logger.error("Error sending unregister service event for " +
                                     self.services[service_id].service_type + " (" +
                                     self.services[service_id].instance_id + "): " + str(ex))
                else:
                    try:
                        self.services[service_id].start_timer()
                    except Exception, ex:
                        logger.error("Failed to start timer thread for service " +
                                     self.services[service_id].service_type + " (" +
                                     self.services[service_id].instance_id + "): " + str(ex))

    def on_disconnect(self):
        """
         On disconnect, send unregister event for all active services as long as still connected.
        Stop all timer threads.
        
        :return: None.
        """
        with self.lock:

            for handler in self.services:
                if self.__client.connected:
                    try:
                        self.services[handler].send_unregister_service_event()
                    except Exception, ex:
                        logger.error("Error sending unregister service event for " +
                                     self.services[handler].service_type + " (" +
                                     self.services[handler].instance_id + "): " + str(ex))
                self.services[handler].stop_timer()



# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2014 McAfee Inc. - All Rights Reserved.
################################################################################

# Run with python -m unittest dxlclient.test.test_broker

import unittest
import socket
from mock import patch, ANY
from nose.tools import raises

from dxlclient import Broker
from dxlclient.exceptions import MalformedBrokerUriException


class BrokerTest(unittest.TestCase):
    def setUp(self):
        self.socket_mock = patch('socket.socket', autospec=True).start()
        self.connection_mock = patch('socket.create_connection').start()
        self.connection_mock.return_value = self.socket_mock
        self.broker = Broker(host_name='localhost')

    def tearDown(self):
        patch.stopall()

    @raises(MalformedBrokerUriException)
    def test_parse_invalid_port(self):
        self.broker._parse("a ; b; [c] ")

    @raises(MalformedBrokerUriException)
    def test_parse_missing_hostname(self):
        self.broker._parse("a ; b;")

    @raises(MalformedBrokerUriException)
    def test_parse_port_out_of_range(self):
        self.broker._parse("a ; 65536; [c] ")

    def test_parse_valid_without_ip_address(self):
        self.broker._parse("a ; 8883; [c] ")
        self.assertTrue(self.broker.unique_id == "a")
        self.assertTrue(self.broker.port == 8883)
        self.assertTrue(self.broker.host_name == "c")

    @raises(MalformedBrokerUriException)
    def test_parse_valid_with_invalid_ip_address(self):
        self.broker._parse("a ; 8883; [c];[d] ")

    def test_parse_valid_with_iPv4_address(self):
        self.broker._parse("a ; 8883; [c];10.0.0.1 ")
        self.assertTrue(self.broker.unique_id == "a")
        self.assertTrue(self.broker.port == 8883)
        self.assertTrue(self.broker.host_name == "c")
        self.assertTrue(self.broker.ip_address == "10.0.0.1")

    def test_parse_valid_with_iPv6_address(self):
        self.broker._parse("a ; 8883; [c];[::1] ")
        self.assertTrue(self.broker.unique_id == "a")
        self.assertTrue(self.broker.port == 8883)
        self.assertTrue(self.broker.host_name == "c")
        self.assertTrue(self.broker.ip_address == "::1")

    def test_attributes(self):
        self.assertEqual(self.broker.unique_id, "")
        self.assertEqual(self.broker.host_name, "localhost")
        self.assertEqual(self.broker.ip_address, None)
        self.assertEqual(self.broker.port, 8883)
        self.assertEqual(self.broker._response_from_ip_address, None)
        self.assertEqual(self.broker._response_time, None)

    def test_connect_tries_first_using_hostname(self):
        self.broker._parse("broker_guid;8883;broker.fake.com;1.2.3.4")
        self.broker._connect_to_broker()
        self.assertEqual(self.broker._response_from_ip_address, False)
        self.assertIsNotNone(self.broker._response_time)
        self.connection_mock.assert_called_with(('broker.fake.com', 8883), timeout=ANY)

    def test_connect_uses_ip_address_when_connect_to_hostname_fails(self):
        self.connection_mock.side_effect = [socket.error, self.socket_mock]
        self.broker._parse("broker_guid;8883;broker.fake.com;1.2.3.4")
        self.broker._connect_to_broker()
        self.assertEquals(self.connection_mock.call_count, 2)
        self.assertEqual(self.broker._response_from_ip_address, True)
        self.assertIsNotNone(self.broker._response_time)
        self.connection_mock.assert_called_with(('1.2.3.4', 8883), timeout=ANY)

    def test_close_once(self):
        self.broker._parse("broker_guid;8883;broker.fake.com;1.2.3.4")
        self.broker._connect_to_broker()
        self.assertEquals(self.socket_mock.close.call_count, 1)

    def test_close_once_when_errors(self):
        self.connection_mock.side_effect = [socket.error, self.socket_mock]
        self.broker._parse("broker_guid;8883;broker.fake.com;1.2.3.4")
        self.broker._connect_to_broker()
        self.assertEquals(self.socket_mock.close.call_count, 1)



# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2014 McAfee Inc. - All Rights Reserved.
################################################################################

# Run with python -m unittest dxlclient.test.test_request_manager

import unittest

import dxlclient.exceptions as exceptions
from dxlclient.callbacks import ResponseCallback
from dxlclient import RequestManager
from dxlclient import Request
from dxlclient import Response
from dxlclient import UuidGenerator


class MockDxlClient(object):
    def __init__(self):
        # The unique identifier of the client
        self.unique_id = UuidGenerator.generate_id_as_string()

    def add_response_callback(self, channel, response_callback):
        pass

    def _send_request(self, request):
        pass


class MockResponseCallback(ResponseCallback):
    def on_response(self, response):
        pass


class RequestManagerTest(unittest.TestCase):
    def setUp(self):
        self.client = MockDxlClient()
        self.rm = RequestManager(self.client)

    def test_current_request(self):
        uid = UuidGenerator.generate_id_as_string()
        self.rm.add_current_request(uid)
        self.assertEquals(1, self.rm.get_current_request_queue_size())
        self.assertTrue(uid in self.rm.current_request_message_ids)
        self.rm.remove_current_request(uid)
        self.assertEquals(0, self.rm.get_current_request_queue_size())
        self.assertFalse(uid in self.rm.current_request_message_ids)

    def test_register_wait_for_response(self):
        request = Request(destination_topic="/test")
        self.rm.register_wait_for_response(request)
        self.assertEquals(1, len(self.rm.sync_wait_message_ids))
        self.assertTrue(request.message_id in self.rm.sync_wait_message_ids)
        self.rm.unregister_wait_for_response(request)
        self.assertEquals(0, len(self.rm.sync_wait_message_ids))
        self.assertFalse(request.message_id in self.rm.sync_wait_message_ids)

    def test_register_async_callback(self):
        request = Request(destination_topic="/test")
        cb = MockResponseCallback()

        self.rm.register_async_callback(request, cb)
        self.assertTrue(request.message_id in self.rm.callback_map)
        self.rm.unregister_async_callback(request.message_id)
        self.assertFalse(request.message_id in self.rm.callback_map)

    def test_wait_for_response(self):
        request = Request(destination_topic="/test")

        with self.assertRaises(exceptions.WaitTimeoutException):
            self.rm.wait_for_response(request, 2)

    def test_on_response_for_sync_request(self):
        request = Request(destination_topic="/test")
        self.rm.register_wait_for_response(request)

        self.assertEquals(1, len(self.rm.sync_wait_message_ids))
        self.assertTrue(request.message_id in self.rm.sync_wait_message_ids)

        response = Response(request=request)
        self.rm.on_response(response)

        self.assertEquals(0, len(self.rm.sync_wait_message_ids))
        self.assertEquals(1, len(self.rm.sync_wait_message_responses))
        self.assertTrue(request.message_id in self.rm.sync_wait_message_responses)

        result = self.rm.wait_for_response(request, 2)

        self.assertEquals(request.message_id, result.request_message_id)

    def test_sync_request(self):
        request = Request(destination_topic="/test")

        with self.assertRaises(exceptions.WaitTimeoutException):
            self.rm.sync_request(request, 2)

        self.assertEquals(0, len(self.rm.sync_wait_message_ids))
        self.assertEquals(0, len(self.rm.sync_wait_message_responses))

    def test_async_request(self):
        request = Request(destination_topic="/test")

        class TestResponseCallback(ResponseCallback):
            def __init__(self):
                super(TestResponseCallback, self).__init__()
                self.response = None

            def on_response(self, resp):
                self.response = resp

        cb = TestResponseCallback()
        self.assertIsNone(cb.response)

        self.rm.async_request(request, cb)

        self.assertEquals(0, len(self.rm.sync_wait_message_ids))
        self.assertEquals(0, len(self.rm.sync_wait_message_responses))
        self.assertTrue(request.message_id in self.rm.callback_map)

        response = Response(request=request)
        self.rm.on_response(response)

        self.assertIsNotNone(cb.response)
        self.assertEquals(request.message_id, cb.response.request_message_id)

        self.assertEquals(0, len(self.rm.sync_wait_message_ids))
        self.assertEquals(0, len(self.rm.sync_wait_message_responses))
        self.assertFalse(request.message_id in self.rm.callback_map)



# This sample demonstrates wrapping an existing service and exposing it on the
# DXL fabric.
#
# In this particular case, the "openweathermap.org" current weather data is
# exposed as a DXL service. This service wrapper delegates to the
# OpenWeatherMap REST API.
#
# openweather_common.py must be edited to include the OpenWeatherMap API
# key (see http://openweathermap.org/appid)

import json
import logging
import os
import sys

from dxlclient.client import DxlClient
from dxlclient.client_config import DxlClientConfig
from dxlclient.message import Message, Request

from openweather_common import *

# Import common logging and configuration
sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/..")
from common import *

# Configure local logger
logging.getLogger().setLevel(logging.INFO)
logger = logging.getLogger(__name__)

# Create DXL configuration from file
config = DxlClientConfig.create_dxl_config_from_file(CONFIG_FILE)

# Create the client
with DxlClient(config) as client:

    # Connect to the fabric
    client.connect()

    #
    # Invoke the service (send a request)
    #

    # Create the "Current Weather" request
    req = Request(SERVICE_CURRENT_WEATHER_TOPIC)
    # Populate the request payload
    # Examples include:
    #   By ZIP code: zip=97140,us
    #   By geographic coordinates: lat=35&lon=139
    #   By city name: q=London,uk
    req.payload = "zip=97140,us".encode()

    # Send the request and wait for a response (synchronous)
    res = client.sync_request(req)

    # Extract information from the response (if an error did not occur)
    if res.message_type != Message.MESSAGE_TYPE_ERROR:
        response_dict = json.loads(res.payload.decode(encoding="UTF-8"))
        print "Client received response payload: \n" + \
          json.dumps(response_dict, sort_keys=True, indent=4, separators=(',', ': '))
    else:
        logger.error("Error: " + res.error_message + " (" + str(res.error_code) + ")")


# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2017 McAfee Inc. - All Rights Reserved.
################################################################################
import threading
import logging
import ssl
import traceback
import random

import paho.mqtt.client as mqtt # pylint: disable=import-error

from dxlclient import _BaseObject
from dxlclient.client_config import DxlClientConfig
import dxlclient._callback_manager as callback_manager
from dxlclient._request_manager import RequestManager
from dxlclient.exceptions import DxlException
from dxlclient.message import Message, Event, Request, Response, ErrorResponse
from dxlclient._thread_pool import ThreadPool
from dxlclient.exceptions import _raise_wrapped_exception
from dxlclient.service import _ServiceManager
from dxlclient._uuid_generator import UuidGenerator
from _dxl_utils import DxlUtils

__all__ = [
    # Callbacks
    "_on_connect", "_on_disconnect", "_on_message", "_on_log",
    # Client
    "DxlClient",
    # Constants
    "DXL_ERR_AGAIN", "DXL_ERR_SUCCESS", "DXL_ERR_INVALID", "DXL_ERR_INTERRUPT",
]


logger = logging.getLogger(__name__)

DXL_ERR_AGAIN = -1
DXL_ERR_SUCCESS = 0
DXL_ERR_INVALID = 1
DXL_ERR_INTERRUPT = 2

################################################################################
#
# Callbacks
#
################################################################################


def _on_connect(client, userdata, rc): # pylint: disable=invalid-name
    """
    Called when the broker responds to our connection request.

    :param client: The Paho MQTT client reference
    :param userdata: The user data object provided
    :param rc: The result code
    :return: None
    """
    if userdata is None or not isinstance(userdata, DxlClient):
        raise ValueError("User data object not specified")

    self = userdata

    with self._connected_lock:
        self._connected = True

        logger.info("Connected with result code %s", str(rc))

        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        with self._subscriptions_lock:
            for subscription in self._subscriptions:
                try:
                    logger.info("Subscribing to %s", subscription)
                    client.subscribe(subscription)
                except Exception as ex:
                    logger.error("Error during subscribe: %s", ex.message)
                    logger.debug(traceback.format_exc())

        if self._service_manager:
            self._service_manager.on_connect()

        # Notify that connection status has changed
        self._connected_wait_condition.notify_all()


def _on_disconnect(client, userdata, rc): # pylint: disable=invalid-name
    """
    Called when the client disconnects from the broker.

    :param client: The Paho MQTT client reference
    :param userdata: The user data object provided
    :param rc: The result code
    :return: None
    """
    # Remove unused parameter
    del client
    # Check userdata; this needs to be an instance of the DxlClient
    if userdata is None or not isinstance(userdata, DxlClient):
        raise ValueError("User data object not specified")
    self = userdata

    with self._connected_lock:
        self._connected = False
        self._connected_wait_condition.notify_all()

    self._reset_current_broker()

    if rc == 0:
        logger.info("Disconnected with result code %s", str(rc))
    else:
        logger.error("Unexpected disconnect with result code %s", str(rc))
        # Disconnect the client without stopping the event loop
        self._disconnect(stop_event_loop=False)
        # Connect the client without starting a new event loop
        # Start the connect thread
        if self.config.reconnect_when_disconnected:
            self._start_connect_thread(start_event_loop=False)

def _on_message(client, userdata, msg): # pylint: disable=invalid-name
    """
    Called when a message has been received on a topic that the client subscribes to.
    self callback will be called for every message received.

    :param client: The Paho MQTT client reference
    :param userdata: The user data object provided
    :param msg: The message object
    :return: None
    """
    # Remove unused parameter
    del client
    # Check userdata; this needs to be an instance of the DxlClient
    if userdata is None or not isinstance(userdata, DxlClient):
        raise ValueError("Client reference not specified")
    self = userdata

    logger.info("Message received for topic %s", msg.topic)

    # TODO: The behavior is one thread per message, but the individual callbacks are handled sequential.
    # TODO: self is the same as in the Java client, but is not ideal.  One plugin could block the whole
    # TODO: execution.

    try:
        self._thread_pool.add_task(self._handle_message, channel=msg.topic, payload=msg.payload)
    except Exception as ex:  # pylint: disable=broad-except
        logger.error("Error handling message: %s", ex.message)
        logger.debug(traceback.format_exc())


def _on_log(client, userdata, level, buf):
    """
    Called when the client has log information. Define to allow debugging.
    The level variable gives the severity of the message and will be one of
    MQTT_LOG_INFO, MQTT_LOG_NOTICE, MQTT_LOG_WARNING, MQTT_LOG_ERR, and
    MQTT_LOG_DEBUG. The message itself is in buf.

    :param client: The Paho MQTT client reference
    :param userdata: The user data object provided
    :param level: The severity of the message
    :param buf: The message itself
    :return: None
    """
    # Remove unused parameter
    del client, userdata

    if level == mqtt.MQTT_LOG_INFO:
        logger.info("MQTT: %s", str(buf))
    elif level == mqtt.MQTT_LOG_NOTICE:
        logger.info("MQTT: %s", str(buf))
    elif level == mqtt.MQTT_LOG_WARNING:
        logger.warn("MQTT: %s", str(buf))
    elif level == mqtt.MQTT_LOG_ERR:
        logger.error("MQTT: %s", str(buf))
    elif level == mqtt.MQTT_LOG_DEBUG:
        logger.debug("MQTT: %s", str(buf))

################################################################################
#
# DxlClient
#
################################################################################


# pylint: disable=too-many-instance-attributes, invalid-name, too-many-public-methods
class DxlClient(_BaseObject):
    """
    The :class:`DxlClient` class is responsible for all communication with the Data Exchange Layer (DXL)
    fabric (it can be thought of as the "main" class). All other classes exist to support the functionality
    provided by the client.

    The following example demonstrates the configuration of a :class:`DxlClient` instance and
    connecting it to the fabric:

    .. code-block:: python

        from dxlclient.broker import Broker
        from dxlclient.client import DxlClient
        from dxlclient.client_config import DxlClientConfig

        # Create the client configuration
        config = DxlClientConfig(
            broker_ca_bundle="c:\\\\certs\\\\brokercerts.crt",
            cert_file="c:\\\\certs\\\\client.crt",
            private_key="c:\\\\certs\\\\client.key",
            brokers=[Broker.parse("ssl://192.168.189.12")])

        # Create the DXL client
        with DxlClient(config) as dxl_client:

            # Connect to the fabric
            dxl_client.connect()

    **NOTE:** The preferred way to construct the client is via the Python "with" statement as shown above. The "with"
    statement ensures that resources associated with the client are properly cleaned up when the block is exited.

    The following modules support the client:

    - :mod:`dxlclient.client_config` : See this module for information on configuring the :class:`DxlClient`
    - :mod:`dxlclient.message` : See this module for information on the different types of messages that can be
      exchanged over the DXL fabric
    - :mod:`dxlclient.callbacks` : See this module for information on registering "callbacks" that are used to
      receive messages via the :class:`DxlClient`. This module also includes an example that demonstrate how to
      send :class:`dxlclient.message.Event` messages.
    - :mod:`dxlclient.service` : See this module for information on registering "services" with the DXL fabric.
      This module also includes an example that demonstrates how to invoke a DXL service via the :class:`DxlClient`.
    """

    # The default "reply-to" prefix. self is typically used for setting up response
    # channels for requests, etc.
    _REPLY_TO_PREFIX = "/mcafee/client/"
    # The default wait time for a synchronous request, defaults to 1 hour
    _DEFAULT_WAIT = 60 * 60
    # The default wait for policy delay (in seconds)
    _DEFAULT_WAIT_FOR_POLICY_DELAY = 2
    # The default minimum amount of threads in a thread pool
    _DEFAULT_MIN_POOL_SIZE = 10
    # The default maximum amount of threads in a thread pool
    _DEFAULT_MAX_POOL_SIZE = 25
    # The default quality of server (QOS) for messages
    _DEFAULT_QOS = 0
    # The default connect wait
    _DEFAULT_CONNECT_WAIT = 10  # seconds

    def __init__(self, config):
        """
        Constructor parameters:

        :param config: The :class:`dxlclient.client_config.DxlClientConfig` object containing the configuration
            settings for the client.
        """
        super(DxlClient, self).__init__()

        if config is None or not isinstance(config, DxlClientConfig):
            raise ValueError("Client configuration not specified")

        # The client configuration
        self._config = config
        # The lock for the client configuration
        self._config_lock = threading.RLock()
        # The condition associated with the client configuration
        self._config_lock_condition = threading.Condition(self._config_lock)

        # The flag for the connection state
        self._connected = False
        # The lock for the flag for the connection state
        self._connected_lock = threading.RLock()
        # The condition for the flag on connection state
        self._connected_wait_condition = threading.Condition(self._connected_lock)
        # The current broker the client is connected to
        self._current_broker = None
        # The lock for the current broker the client is connected to
        self._current_broker_lock = threading.RLock()

        # The default wait time for a synchronous request
        self._default_wait = self._DEFAULT_WAIT

        # The wait for policy delay (in seconds)
        self._wait_for_policy_delay = self._DEFAULT_WAIT_FOR_POLICY_DELAY

        # The minimum amount of threads in a thread pool
        self._core_pool_size = self._DEFAULT_MIN_POOL_SIZE
        # The maximum amount of threads in a thread pool
        self._maximum_pool_size = self._DEFAULT_MAX_POOL_SIZE

        # The quality of server (QOS) for messages
        self._qos = self._DEFAULT_QOS
        # The "reply-to" prefix. self is typically used for setting up response
        # channels for requests, etc.
        self._reply_to_topic = self._REPLY_TO_PREFIX + self._config._client_id

        # The request callbacks manager
        self._request_callbacks = callback_manager._RequestCallbackManager()
        # The response callbacks manager
        self._response_callbacks = callback_manager._ResponseCallbackManager()
        # The event callbacks manager
        self._event_callbacks = callback_manager._EventCallbackManager()

        # The current list of subscriptions
        self._subscriptions = set()
        # The lock for the current list of subscriptions
        self._subscriptions_lock = threading.RLock()

        # The underlying MQTT client instance
        self._client = mqtt.Client(client_id=self._config._client_id,
                                  clean_session=True,
                                  userdata=self,
                                  protocol=mqtt.MQTTv31)

        # The MQTT client connect callback
        self._client.on_connect = _on_connect
        # The MQTT client disconnect callback
        self._client.on_disconnect = _on_disconnect
        # The MQTT client message callback
        self._client.on_message = _on_message
        # The MQTT client log callback
        if logger.isEnabledFor(logging.DEBUG):
            self._client.on_log = _on_log

        # pylint: disable=no-member
        # The MQTT client TLS configuration
        self._client.tls_set(config.broker_ca_bundle,
                            certfile=config.cert_file,
                            keyfile=config.private_key,
                            cert_reqs=ssl.CERT_REQUIRED,
                            tls_version=ssl.PROTOCOL_SSLv23,
                            ciphers=None)
        # The MQTT client TLS configuration to bypass hostname validation
        self._client.tls_insecure_set(True)

        # Generate a message pool prefix
        self._message_pool_prefix = "DxlMessagePool-" + UuidGenerator.generate_id_as_string()

        # The thread pool for message handling
        self._thread_pool = ThreadPool(
            num_threads=config.incoming_message_thread_pool_size,
            queue_size=config.incoming_message_queue_size,
            thread_prefix = self._message_pool_prefix)

        # Subscribe to the client reply channel
        self.subscribe(self._reply_to_topic)

        # The request manager (manages synchronous and asynchronous request callbacks,
        # notifications, etc.).
        self._request_manager = RequestManager(client=self)

        # The service manager (manages services request callbacks, notifications, etc.).
        self._service_manager = _ServiceManager(client=self)

        # The loop thread
        self._thread = None
        # The loop thread terminate flag
        self._thread_terminate = False

        # The lock for the connect thread
        self._connect_wait_lock = threading.RLock()
        # The condition associated with the client configuration
        self._connect_wait_condition = threading.Condition(self._connect_wait_lock)

        self._destroy_lock = threading.RLock()
        self._destroyed = False

    def __del__(self):
        """destructor"""
        super(DxlClient, self).__del__()
        self.destroy()

    def __enter__(self):
        """Enter with"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit with"""
        self.destroy()

    @property
    def config(self):
        """
        The :class:`dxlclient.client_config.DxlClientConfig` instance that was specified when the
        client was constructed.

        See :mod:`dxlclient.client_config` for more information on configuring the client.
        """
        with self._config_lock:
            return self._config

    @property
    def connected(self):
        """Whether the client is currently connected to the DXL fabric."""
        with self._connected_lock:
                return self._connected

    def connect(self):
        """
        Attempts to connect the client to the DXL fabric.

        This method does not return until either the client has connected to the fabric or it has exhausted
        the number of retries configured for the client causing an exception to be raised.

        Several attributes are available for controlling the client retry behavior:

        - :attr:`dxlclient.client_config.DxlClientConfig.connect_retries` : The maximum number of connection attempts
          for each :class:`dxlclient.broker.Broker` specified in the :class:`dxlclient.client_config.DxlClientConfig`
        - :attr:`dxlclient.client_config.DxlClientConfig.reconnect_delay` : The initial delay between retry attempts.
          The delay increases ("backs off") as subsequent connection attempts are made.
        - :attr:`dxlclient.client_config.DxlClientConfig.reconnect_back_off_multiplier` : Multiples the current
          reconnect delay by this value on subsequent connect retries. For example, a current delay of 3 seconds
          with a multiplier of 2 would result in the next retry attempt being in 6 seconds.
        - :attr:`dxlclient.client_config.DxlClientConfig.reconnect_delay_random` : A randomness delay percentage
          (between 0.0 and 1.0) that is used to increase the current retry delay by a random amount for the purpose
          of preventing multiple clients from having the same retry pattern
        - :attr:`dxlclient.client_config.DxlClientConfig.reconnect_delay_max` : The maximum delay between retry attempts
        """
        if self.connected:
            raise DxlException("Already connected")

        if self._thread is not None:
            raise DxlException("Already trying to connect")

        # Start the connect thread
        self._start_connect_thread(connect_retries=self.config.connect_retries)

        # Wait for the connect thread to finish
        if self._thread is not None:
            while self._thread.isAlive():
                self._thread.join(1)
            self._thread = None

        # Wait for the callback to be invoked
        with self._connected_lock:
            if not self.connected:
                self._connected_wait_condition.wait(5)

        # Check if we were connected
        if not self.connected:
            raise DxlException("Failed to establish connection")

    def _start_connect_thread(self, start_event_loop=True, connect_retries=-1):
        self._thread = threading.Thread(target=self._connect_thread_main, args=[start_event_loop, connect_retries])
        self._thread.daemon = True
        self._thread.start()

    def destroy(self):
        """
        Destroys the client (releases all associated resources).

        **NOTE:** Once the method has been invoked, no other calls should be made to the client.

        Also note that this method should rarely be called directly. Instead, the preferred usage of the
        client is via a Python "with" statement as shown below:

        .. code-block:: python

            # Create the DXL client
            with DxlClient(config) as dxl_client:

                # Connect to the fabric
                dxl_client.connect()

        The "with" statement ensures that resources associated with the client are properly cleaned up when the block
        is exited (the :func:`destroy` method is invoked).

        """
        with self._destroy_lock:
            if not self._destroyed:
                self._service_manager.destroy()
                self._service_manager = None

                self._request_manager.destroy()
                self._request_manager = None

                self.disconnect()

                self._thread_pool.shutdown()

                self._config = None

                self._client.user_data_set( None )
                self._client = None

                self._destroyed = True

    def disconnect(self):
        """
        Attempts to disconnect the client from the DXL fabric.
        """
        if self._connected:
            self._disconnect(stop_event_loop=True)
        else:
            logger.warning("Trying to disconnect a disconnected client.")

    def _disconnect(self, stop_event_loop=False):

        if self._service_manager:
            self._service_manager.on_disconnect()

        logger.info("Waiting for thread pool completion...")
        self._thread_pool.wait_completion()

        for subscription in self._subscriptions:
            if self.connected:
                try:
                    logger.info("Unsubscribing from %s", subscription)
                    self._client.unsubscribe(subscription)
                except Exception as ex:  # pylint: disable=broad-except
                    logger.error("Error during unsubscribe: %s", ex.message)
                    logger.debug(traceback.format_exc())

        # In case of a reconnect after connection loss, the event loop will
        # not be stopped and the client will not be forcefully disconnected.
        if stop_event_loop:
            logger.info("Stopping event loop...")
            self._client.loop_stop()
            logger.info("Trying to disconnect client...")
            self._client.disconnect()
            logger.info("Disconnected.")

        # Make sure the connection loop is done
        if self._thread is not None:
            logger.info("Waiting for the thread to terminate...")
            self._thread_terminate = True
            with self._connect_wait_lock:
                self._connect_wait_condition.notifyAll()
            while self._thread.isAlive():
                self._thread.join(1)
            self._thread = None
            logger.info("Thread terminated.")

    def _connect_thread_main(self, start_loop, connect_retries):
        """
        The connection thread main function
        """
        self._thread_terminate = False
        self._loop_until_connected(start_loop, connect_retries)

    def _connect(self, brokers):
        """
        Internal function that attempts to connect to one of the given brokers.

        :param brokers: The (sorted) list of brokers
        """
        self._reset_current_broker()
        keep_alive_interval = self.config.keep_alive_interval
        latest_ex = None

        for broker in brokers:
            if self._thread_terminate:
                break
            if broker._response_time is not None:
                try:
                    if broker._response_from_ip_address:
                        logger.info("Trying to connect to broker %s...", broker.to_string())
                        self._client.connect(broker.ip_address, broker.port, keep_alive_interval)
                    else:
                        logger.info("Trying to connect to broker %s...", broker.to_string())
                        self._client.connect(broker.host_name, broker.port, keep_alive_interval)
                    self._current_broker = broker
                    break
                except Exception as ex:  # pylint: disable=broad-except
                    logger.error("Failed to connect to broker %s: %s",
                                 broker.to_string(),
                                 ex.message)
                    logger.debug(traceback.format_exc())
                    latest_ex = ex

        if self._current_broker is None:
            for broker in brokers:
                if self._thread_terminate:
                    break
                try:
                    logger.info(
                        "Trying to connect to broker (host name) %s...", broker.to_string())
                    self._client.connect(broker.host_name, broker.port, keep_alive_interval)
                    self._current_broker = broker
                    break
                except Exception as ex:  # pylint: disable=broad-except
                    logger.error("Failed to connect to broker (host name) %s: %s",
                                 broker.to_string(), str(ex))
                    logger.debug(traceback.format_exc())
                    latest_ex = ex

                if self._thread_terminate:
                    break
                if self._current_broker is None and broker.ip_address is not None:
                    try:
                        logger.info(
                            "Trying to connect to broker (IP address) %s (%s:%d)...",
                            broker.unique_id, broker.ip_address, broker.port)
                        self._client.connect(broker.ip_address, broker.port, keep_alive_interval)
                        self._current_broker = broker
                        break
                    except Exception as ex:  # pylint: disable=broad-except
                        logger.error("Failed to connect to broker (IP address) %s: %s",
                                     broker.to_string(), str(ex))
                        logger.debug(traceback.format_exc())
                        latest_ex = ex

        if self._current_broker is not None:
            logger.info("Connected to broker %s",
                        self._current_broker.unique_id)
        else:
            if latest_ex is not None:
                raise latest_ex  # pylint: disable=raising-bad-type

    def _loop_until_connected(self, start_loop, connect_retries):

        # The client is already connected
        if self.connected:
            logger.error("Already connected")
            return DXL_ERR_INVALID

        logger.info("Waiting for broker list...")
        self._config_lock.acquire()
        try:
            while not self._thread_terminate and len(self._config.brokers) == 0:
                self._config_lock_condition.wait(self._wait_for_policy_delay)
                if len(self._config.brokers) == 0:
                    logger.debug("No broker defined. Waiting for broker list...")
        finally:
            self._config_lock.release()

        if self._thread_terminate is True:
            logger.debug("Stopping...")
            return DXL_ERR_INTERRUPT

        logger.info("Checking brokers...")
        brokers = self._config._get_sorted_broker_list()

        logger.info("Trying to connect...")
        retries = connect_retries
        retry_delay = self.config.reconnect_delay
        first_attempt = True
        latest_ex = None
        latest_ex_traceback = None

        while not self._thread_terminate and (connect_retries < 0 or retries >= 0):
            if not first_attempt:
                # Determine retry delay
                retry_delay_max = self.config.reconnect_delay_max
                if retry_delay > retry_delay_max:
                    retry_delay = retry_delay_max
                # Apply random after max (so we still have randomness, may exceed maximum)
                retry_delay += ((self.config.reconnect_delay_random * retry_delay) * random.random())

                logger.error("Retrying connect in %s seconds: %s", str(retry_delay), str(latest_ex))

                # Wait...
                with self._connect_wait_lock:
                    self._connect_wait_condition.wait(retry_delay)

                # Update retry delay
                retry_delay *= self.config.reconnect_back_off_multiplier

            try:
                self._connect(brokers)
                break
            except Exception as ex:
                # Track latest exception
                latest_ex = ex
                latest_ex_traceback = traceback.format_exc()

            first_attempt = False
            retries -= 1

        if self._thread_terminate is True:
            logger.info("Stopping...")
            return DXL_ERR_INTERRUPT

        if not self._current_broker:
            if latest_ex:
                logger.error("Error during connect: %s", latest_ex.message)
            if latest_ex_traceback:
                logger.debug(latest_ex_traceback)

        logger.info("Launching event loop...")

        if start_loop:
            self._client.loop_start()

        return DXL_ERR_SUCCESS

    @property
    def current_broker(self):
        """
        The :class:`dxlclient.broker.Broker` that the client is currently connected to. ``None`` is returned
        if the client is not currently connected to a :class:`dxlclient.broker.Broker`.
        """
        with self._current_broker_lock:
            if self.connected:
                return self._current_broker
            else:
                return None

    def _set_current_broker(self, current_broker):
        """
        Internal method. Sets the current broker.

        :param current_broker: {@code dxlclient.broker.Broker} to set as current broker.
        """
        with self._current_broker_lock:
            self._current_broker = current_broker

    def _reset_current_broker(self):
        """
        Clean current broker.
        """
        with self._current_broker_lock:
            self._current_broker = None

    def subscribe(self, topic):
        """
        Subscribes to the specified topic on the DXL fabric. This method is typically used in
        conjunction with the registration of :class:`dxlclient.callbacks.EventCallback` instances
        via the :func:`add_event_callback` method.

        The following is a simple example of using this:

        .. code-block:: python

            from dxlclient.callbacks import EventCallback

            class MyEventCallback(EventCallback):
                def on_event(self, event):
                    print "Received event! " + event.source_client_id

            dxl_client.add_event_callback("/testeventtopic", MyEventCallback(), False)
            dxl_client.subscribe("/testeventtopic")

        **NOTE:** By default when registering an event callback the client will automatically subscribe to the topic.
        In this example the :func:`dxlclient.client.DxlClient.add_event_callback` method is invoked with the
        ``subscribe_to_topic`` parameter set to ``False`` preventing the automatic subscription.

        :param topic: The topic to subscribe to
        """
        logger.debug("%s(): Waiting for Subscriptions lock...", DxlUtils.func_name())
        self._subscriptions_lock.acquire()
        try:
            if topic not in self._subscriptions:
                self._subscriptions.add(topic)
                if self.connected:
                    self._client.subscribe(topic)
        except Exception as ex:
            logger.error("Error during subscribe: %s", ex.message)
            logger.debug(traceback.format_exc())
            raise DxlException("Error during subscribe" + str(ex))
        finally:
            logger.debug("%s(): Releasing Subscriptions lock.", DxlUtils.func_name())
            self._subscriptions_lock.release()

    def unsubscribe(self, topic):
        """
        Unsubscribes from the specified topic on the DXL fabric.

        See the :func:`subscribe` method for more information on subscriptions.

        :param topic: The topic to unsubscribe from
        """
        logger.debug("%s(): Waiting for Subscriptions lock...", DxlUtils.func_name())
        self._subscriptions_lock.acquire()
        try:
            if topic in self._subscriptions:
                if self.connected:
                    self._client.unsubscribe(topic)
        except Exception as ex:
            logger.error("Error during unsubscribe: %s", ex.message)
            logger.debug(traceback.format_exc())
            raise DxlException("Error during unsubscribe")
        finally:
            self._subscriptions.remove(topic)
            logger.debug("%s(): Releasing Subscriptions lock.", DxlUtils.func_name())
            self._subscriptions_lock.release()

    @property
    def subscriptions(self):
        """
        A tuple containing the topics that the client is currently subscribed to

        See :func:`subscribe` for more information on adding subscriptions
        """
        logger.debug("%s(): Waiting for Subscriptions lock...", DxlUtils.func_name())
        self._subscriptions_lock.acquire()
        try:
            return tuple(self._subscriptions)
        finally:
            logger.debug("%s(): Releasing Subscriptions lock.", DxlUtils.func_name())
            self._subscriptions_lock.release()

    def sync_request(self, request, timeout=_DEFAULT_WAIT):
        """
        Sends a :class:`dxlclient.message.Request` message to a remote DXL service.

        See module :mod:`dxlclient.service` for more information on DXL services.

        :param request: The :class:`dxlclient.message.Request` message to send to a remote DXL service
        :param timeout: The amount of time (in seconds) to wait for the :class:`dxlclient.message.Response`
            to the request. If the timeout is exceeded an exception will be raised. Defaults to ``3600``
            seconds (1 hour)
        """
        if threading.currentThread().name.startswith(self._message_pool_prefix):
            raise DxlException("Synchronous requests may not be invoked while handling an incoming message. " +
                               "The synchronous request must be made on a different thread.")

        return self._request_manager.sync_request(request, timeout)

    def async_request(self, request, response_callback=None):
        """
        Sends a :class:`dxlclient.message.Request` message to a remote DXL service asynchronously.
        This method differs from :func:`sync_request` due to the fact that it returns to the caller
        immediately after delivering the :class:`dxlclient.message.Request` message to the DXL fabric (It does
        not wait for the corresponding :class:`dxlclient.message.Response` to be received).

        An optional :class:`dxlclient.callbacks.ResponseCallback` can be specified. This callback will be invoked
        when the corresponding :class:`dxlclient.message.Response` message is received by the client.

        See module :mod:`dxlclient.service` for more information on DXL services.

        :param request: The :class:`dxlclient.message.Request` message to send to a remote DXL service
        :param response_callback: An optional :class:`dxlclient.callbacks.ResponseCallback` that will be invoked
            when the corresponding :class:`dxlclient.message.Response` message is received by the client.
        """
        return self._request_manager.async_request(request, response_callback)

    def _get_async_callback_count(self):
        """
        Returns the count of async callbacks that are waiting for a response
        :return: The count of async callbacks that are waiting for a response
        """
        return self._request_manager._get_async_callback_count()

    def _publish_message(self, channel, payload, qos):
        """
        Publishes the specified message

        :param channel: The channel to publish on
        :param payload: The message content
        :param qos: The quality of service (QOS)
        """
        try:
            self._client.publish(topic=channel, payload=payload, qos=qos)
        except Exception as ex:  # pylint: disable=broad-except
            logger.error("Error publishing message: %s", ex.message)
            logger.debug(traceback.format_exc())
            _raise_wrapped_exception("Error publishing message", ex)

    def _send_request(self, request):
        """
        Sends the specified request to the DXL fabric.

        :param request: The request to send to the DXL fabric
        """
        if request is None or not isinstance(request, Request):
            raise ValueError("Invalid or unspecified request object")
        request.reply_to_topic = self._reply_to_topic
        self._publish_message(request.destination_topic, request._to_bytes(), self._qos)

    def send_response(self, response):
        """
        Attempts to deliver the specified :class:`dxlclient.message.Response` message to the DXL fabric.
        The fabric will in turn attempt to deliver the response back to the client who sent the
        corresponding :class:`dxlclient.message.Request`.

        See module :mod:`dxlclient.message` for more information on message types, how they are delivered to
        remote clients, etc.

        See module :mod:`dxlclient.service` for more information on DXL services.

        :param event: The :class:`dxlclient.message.Event` to send
        """
        if response is None or not isinstance(response, Response):
            raise ValueError("Invalid or unspecified response object")
        self._publish_message(response.destination_topic, response._to_bytes(), self._qos)

    def send_event(self, event):
        """
        Attempts to deliver the specified :class:`dxlclient.message.Event` message to the DXL fabric.

        See module :mod:`dxlclient.message` for more information on message types, how they are delivered to
        remote clients, etc.

        :param event: The :class:`dxlclient.message.Event` to send
        """
        if event is None or not isinstance(event, Event):
            raise ValueError("Invalid or unspecified event object")
        self._publish_message(event.destination_topic, event._to_bytes(), self._qos)

    def add_request_callback(self, topic, request_callback):
        """
        Adds a :class:`dxlclient.callbacks.RequestCallback` to the client for the specified topic.
        The callback will be invoked when :class:`dxlclient.message.Request` messages are received by the client
        on the specified topic. A topic of ``None`` indicates that the callback should receive
        :class:`dxlclient.message.Request` messages for all topics (no filtering).

        **NOTE:** Usage of this method is quite rare due to the fact that registration of
        :class:`dxlclient.callbacks.RequestCallback` instances with the client occurs automatically when
        registering a service. See module :mod:`dxlclient.service` for more information on DXL services.

        :param topic: The topic to receive :class:`dxlclient.message.Request` messages on. A topic of ``None`` indicates
            that the callback should receive :class:`dxlclient.message.Request` messages for all topics (no filtering).
        :param request_callback: The :class:`dxlclient.callbacks.RequestCallback` to be invoked when a
            :class:`dxlclient.message.Request` message is received on the specified topic
        """
        self._request_callbacks.add_callback(("" if topic is None else topic), request_callback)

    def remove_request_callback(self, topic, request_callback):
        """
        Removes a :class:`dxlclient.callbacks.RequestCallback` from the client for the specified topic. This method
        must be invoked with the same arguments as when the callback was originally registered via
        :func:`add_request_callback`.

        :param topic: The topic to remove the callback for
        :param request_callback: The :class:`dxlclient.callbacks.RequestCallback` to be removed for the specified topic
        """
        self._request_callbacks.remove_callback(("" if topic is None else topic), request_callback)

    def add_response_callback(self, topic, response_callback):
        """
        Adds a :class:`dxlclient.callbacks.ResponseCallback` to the client for the specified topic.
        The callback will be invoked when :class:`dxlclient.message.Response` messages are received by the client
        on the specified topic. A topic of ``None`` indicates that the callback should receive
        :class:`dxlclient.message.Response` messages for all topics (no filtering).

        **NOTE:** Usage of this method is quite rare due to the fact that the use of
        :class:`dxlclient.callbacks.ResponseCallback` instances are typically limited to invoking a
        remote DXL service via the :func:`async_request` method.

        :param topic: The topic to receive :class:`dxlclient.message.Request` messages on. A topic of ``None`` indicates
            that the callback should receive :class:`dxlclient.message.Request` messages for all topics (no filtering).
        :param request_callback: The :class:`dxlclient.callbacks.RequestCallback` to be invoked when a
            :class:`dxlclient.message.Request` message is received on the specified topic
        """
        self._response_callbacks.add_callback(("" if topic is None else topic), response_callback)

    def remove_response_callback(self, topic, response_callback):
        """
        Removes a :class:`dxlclient.callbacks.ResponseCallback` from the client for the specified topic. This method
        must be invoked with the same arguments as when the callback was originally registered via
        :func:`add_response_callback`.

        :param topic: The topic to remove the callback for
        :param response_callback: The :class:`dxlclient.callbacks.ResponseCallback` to be removed for the specified topic
        """
        self._response_callbacks.remove_callback(("" if topic is None else topic), response_callback)

    def add_event_callback(self, topic, event_callback, subscribe_to_topic=True):
        """
        Adds a :class:`dxlclient.callbacks.EventCallback` to the client for the specified topic.
        The callback will be invoked when :class:`dxlclient.message.Event` messages are received by the client
        on the specified topic. A topic of ``None`` indicates that the callback should receive
        :class:`dxlclient.message.Event` messages for all topics (no filtering).

        :param topic: The topic to receive :class:`dxlclient.message.Event` messages on. A topic of ``None`` indicates
            that the callback should receive :class:`dxlclient.message.Event` messages for all topics (no filtering).
        :param event_callback: The :class:`dxlclient.callbacks.EventCallback` to be invoked when a
            :class:`dxlclient.message.Event` message is received on the specified topic
        :param subscribe_to_topic: Optional parameter to indicate if the client should subscribe
            (:func:`dxlclient.client.DxlClient.subscribe`) to the topic.
            By default the client will subscribe to the topic. Specify ``False`` to prevent subscribing to the topic.
        """
        self._event_callbacks.add_callback(("" if topic is None else topic), event_callback)
        if subscribe_to_topic is True and topic is not None:
            self.subscribe(topic)

    def remove_event_callback(self, topic, event_callback, unsubscribe_from_topic=True):
        """
        Removes a :class:`dxlclient.callbacks.EventCallback` from the client for the specified topic. This method
        must be invoked with the same arguments as when the callback was originally registered via
        :func:`add_event_callback`.

        :param topic: The topic to remove the callback for
        :param event_callback: The :class:`dxlclient.callbacks.EventCallback` to be removed for the specified topic
        :param unsubscribe_from_topic: Optional parameter to indicate if the client should also unsubscribe
            (:func:`dxlclient.client.DxlClient.unsubscribe`) from the topic. By default the client will unsubscribe
            from the topic. Specify ``False`` to prevent unsubscribing to the topic.
        """
        self._event_callbacks.remove_callback(("" if topic is None else topic), event_callback)
        if unsubscribe_from_topic is True and topic is not None:
            self.unsubscribe(topic)

    def _fire_request(self, request):
        """
        Fires the specified {@link Request} to {@link RequestCallback} listeners currently
        registered with the client.

        :param request: The {@link Request} to fire
        """
        self._request_callbacks.fire_message(request)

    def _fire_response(self, response):
        """
        Fires the specified {@link Response} to {@link ResponseCallback} listeners currently
        registered with the client.

        :param response: The {@link Response} to fire
        """
        self._response_callbacks.fire_message(response)

    def _fire_event(self, event):
        """
        Fires the specified {@link Event} to {@link EventCallback} listeners currently
        registered with the client.

        :param event: The {@link Event} to fire
        """
        self._event_callbacks.fire_message(event)

    def _handle_message(self, channel, payload):
        """
        Processes an incoming message. The bytes from the message are converted into the appropriate
        message type instance (request, response, event, etc.) and then the corresponding registered
        message callbacks are notified.

        :param channel: The channel that the message arrived on
        :param payload: The message received from the channel (as bytes)
        """
        message = Message._from_bytes(payload)
        message.destination_topic = channel

        if isinstance(message, Event):
            self._fire_event(message)
        elif isinstance(message, Request):
            self._fire_request(message)
        elif isinstance(message, Response) or isinstance(message, ErrorResponse):
            self._fire_response(message)
        else:
            raise ValueError("Unknown message type")

    def register_service_async(self, service_reg_info):
        """
        Registers a DXL service with the fabric asynchronously. The specified
        :class:`dxlclient.service.ServiceRegistrationInfo` instance contains information about the
        service that is to be registered.

        This method differs from :func:`register_service_sync` due to the fact that it returns to the caller
        immediately after sending the registration message to the DXL fabric (It does
        not wait for registration confirmation before returning).

        See :mod:`dxlclient.service` for more information on DXL services.

        :param service_reg_info: A :class:`dxlclient.service.ServiceRegistrationInfo` instance containing information
            about the service that is to be registered.
        """
        if self._service_manager: self._service_manager.add_service(service_reg_info)

    def unregister_service_async(self, service_reg_info):
        """
        Unregisters (removes) a DXL service with from the fabric asynchronously. The specified
        :class:`dxlclient.service.ServiceRegistrationInfo` instance contains information about the
        service that is to be removed.

        This method differs from :func:`unregister_service_sync` due to the fact that it returns to the caller
        immediately after sending the unregistration message to the DXL fabric (It does
        not wait for unregistration confirmation before returning).

        See :mod:`dxlclient.service` for more information on DXL services.

        :param service_reg_info: A :class:`dxlclient.service.ServiceRegistrationInfo` instance containing information
            about the service that is to be unregistered.
        """
        if self._service_manager: self._service_manager.remove_service(service_reg_info.service_id)

    def register_service_sync(self, service_req_info, timeout):
        """
        Registers a DXL service with the fabric. The specified
        :class:`dxlclient.service.ServiceRegistrationInfo` instance contains information about the
        service that is to be registered.

        This method will wait for confirmation of the service registration for up to the specified timeout
        in seconds. If the timeout is exceeded an exception will be raised.

        See :mod:`dxlclient.service` for more information on DXL services.

        :param service_reg_info: A :class:`dxlclient.service.ServiceRegistrationInfo` instance containing information
            about the service that is to be registered.
        :param timeout: The amount of time (in seconds) to wait for confirmation of the service registration.
            If the timeout is exceeded an exception will be raised.
        """
        if self._service_manager:
            if not self.connected:
                raise DxlException("Client is not currently connected")
            self._service_manager.add_service(service_req_info)
            service_req_info._wait_for_registration(timeout=timeout)

    def unregister_service_sync(self, service_req_info, timeout):
        """
        Unregisters (removes) a DXL service from the fabric. The specified
        :class:`dxlclient.service.ServiceRegistrationInfo` instance contains information about the
        service that is to be removed.

        This method will wait for confirmation of the service unregistration for up to the specified timeout
        in seconds. If the timeout is exceeded an exception will be raised.

        See :mod:`dxlclient.service` for more information on DXL services.

        :param service_reg_info: A :class:`dxlclient.service.ServiceRegistrationInfo` instance containing information
            about the service that is to be removed.
        :param timeout: The amount of time (in seconds) to wait for confirmation of the service unregistration.
            If the timeout is exceeded an exception will be raised.
        """
        if self._service_manager:
            if not self.connected:
                raise DxlException("Client is not currently connected")

            if not service_req_info:
                raise ValueError("Undefined service object")

            self._service_manager.remove_service(service_req_info.service_id)
            service_req_info._wait_for_unregistration(timeout=timeout)


# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2017 McAfee Inc. - All Rights Reserved.
################################################################################
import logging
import threading
from Queue import Queue
import os.path as path

from ConfigParser import ConfigParser, NoSectionError, NoOptionError

from dxlclient import _BaseObject, _ObjectTracker
from dxlclient.broker import Broker
from dxlclient._uuid_generator import UuidGenerator
from dxlclient.exceptions import BrokerListError

################################################################################
#
# Static functions
#
################################################################################
logger = logging.getLogger(__name__)


class _DxlConfigParser(ConfigParser):
    """
    DxlClientConfig config file parser
    """

    def __init__(self):
        ConfigParser.__init__(self)
        _ObjectTracker.get_instance().obj_constructed(self)

    def __del__(self):
        _ObjectTracker.get_instance().obj_destructed(self)

    def get(self, section, option, show_warning=True):  # pylint: disable=arguments-differ
        """
        Gets the value of an option, if it doesn't exist return an empty string.

        :param section: Config section string
        :param option: Config option string
        :param show_warning: Show warning log message if option cannot be found in config file
        :return: Option value
        """
        option_value = ''
        try:
            option_value = ConfigParser.get(self, section, option)
        except (NoOptionError, NoSectionError):
            if show_warning:
                logger.warning(option + " was not defined in config file")
        return option_value


def _get_brokers_from_list(broker_list):
    """
    Helper function that generates a broker list from a dictionary containing brokers in this format
    {
        "{guid}": "{guid};port;broker-name;broker_ip",
        "{guid-2}": "{guid-2};port;broker-2-name;broker-2-ip",
        ...
    }
    :param broker_list: Dictionary with brokers info
    :return: list of broker objects
    """
    brokers = None
    if broker_list:
        brokers = []
        for _, value in broker_list.items():
            # Set hostname to avoid validation error, will be overridden by parse
            broker = Broker(host_name='none')
            broker._parse(value)
            brokers.append(broker)
    return brokers


def _get_brokers(broker_list_json):
    """
    From a json object creates a broker list

    :param broker_list_json: json object containing a dictionary with a BrokerList tag
    :return: list of broker objects
    """
    try:
        return _get_brokers_from_list(broker_list_json)
    except Exception, broker_error:
        raise BrokerListError("Broker list is not a valid JSON: " + str(broker_error))

################################################################################
#
# DxlClientConfig
#
################################################################################


class DxlClientConfig(_BaseObject):
    """
    The Data Exchange Layer (DXL) client configuration contains the information necessary to connect
    a :class:`dxlclient.client.DxlClient` to the DXL fabric.

    The configuration includes the required PKI information (client certificate, client private key,
    broker CA certificates) and the set of DXL message brokers that are available to connect to on the fabric.

    The following sample shows creating a client configuration, instantiating a DXL client, and connecting
    to the fabric:

    .. code-block:: python

        from dxlclient.broker import Broker
        from dxlclient.client import DxlClient
        from dxlclient.client_config import DxlClientConfig

        # Create the client configuration
        config = DxlClientConfig(
            broker_ca_bundle="c:\\\\certs\\\\brokercerts.crt",
            cert_file="c:\\\\certs\\\\client.crt",
            private_key="c:\\\\certs\\\\client.key",
            brokers=[Broker.parse("ssl://192.168.189.12")])

        # Create the DXL client
        with DxlClient(config) as dxl_client:

            # Connect to the fabric
            dxl_client.connect()
    """

    # The default number of times to retry during connect, default -1 (infinite)
    _DEFAULT_CONNECT_RETRIES = -1
    # The default keep alive interval (in seconds); client pings broker at interval
    # 30 minutes by default
    _DEFAULT_MQTT_KEEP_ALIVE_INTERVAL = 30 * 60
    # The default reconnect back off multiplier
    _DEFAULT_RECONNECT_BACK_OFF_MULTIPLIER = 2
    # The default reconnect delay (in seconds)
    _DEFAULT_RECONNECT_DELAY = 1
    # The default maximum reconnect delay, defaults to 1 minute
    _DEFAULT_RECONNECT_DELAY_MAX = 60
    # The default reconnect delay random multiplier, defaults to 25 percent
    _DEFAULT_RECONNECT_DELAY_RANDOM = 0.25
    # Whether to attempt to reconnect when disconnected
    _DEFAULT_RECONNECT_WHEN_DISCONNECTED = True

    def __init__(self, broker_ca_bundle, cert_file, private_key, brokers):
        """
        Constructor parameters:

        :param broker_ca_bundle: The file name of a bundle containing the broker CA certificates in PEM format
        :param cert_file: The file name of the client certificate in PEM format
        :param private_key: The file name of the client private key in PEM format
        :param brokers: A list of :class:`dxlclient.broker.Broker` objects representing brokers comprising the
            DXL fabric. When invoking the :func:`dxlclient.client.DxlClient.connect` method, the
            :class:`dxlclient.client.DxlClient` will attempt to connect to the closest broker.
        """
        super(DxlClientConfig, self).__init__()

        client_id = UuidGenerator.generate_id_as_string()

        if not broker_ca_bundle:
            raise ValueError("Broker CA bundle not specified")

        if not cert_file:
            raise ValueError("Certificate file not specified")

        if not private_key:
            raise ValueError("Private key file not specified")

        if brokers is None:
            raise ValueError("Brokers were not specified")

        # The number of times to retry during connect
        self._connect_retries = self._DEFAULT_CONNECT_RETRIES
        # The keep alive interval
        self._keep_alive_interval = self._DEFAULT_MQTT_KEEP_ALIVE_INTERVAL
        # The reconnect back off multiplier
        self._reconnect_back_off_multiplier = self._DEFAULT_RECONNECT_BACK_OFF_MULTIPLIER
        # The reconnect delay (in seconds)
        self._reconnect_delay = self._DEFAULT_RECONNECT_DELAY
        # The maximum reconnect delay
        self._reconnect_delay_max = self._DEFAULT_RECONNECT_DELAY_MAX
        # The reconnect delay random
        self._reconnect_delay_random = self._DEFAULT_RECONNECT_DELAY_RANDOM
        # Whether to reconnect when disconnected
        self._reconnect_when_disconnected = self._DEFAULT_RECONNECT_WHEN_DISCONNECTED

        # The unique identifier of the client
        self._client_id = client_id
        # The list of brokers
        self.brokers = brokers
        # The filename of the CA bundle file in PEM format
        self.broker_ca_bundle = broker_ca_bundle
        # The filename of the client certificate in PEM format (must not have a password)
        self.cert_file = cert_file
        # The filename of the private key used to request the certificates
        self.private_key = private_key
        # Queue for getting the sorted broker list
        self._queue = None
        # The incoming message queue size
        self._incoming_message_queue_size = 1000
        # The incoming thread pool size
        self._incoming_message_thread_pool_size = 1

    def __del__(self):
        """destructor"""
        super(DxlClientConfig, self).__del__()

    @property
    def broker_ca_bundle(self):
        """
        The file name of a bundle containing the broker CA certificates in PEM format
        """
        return self._broker_ca_bundle

    @broker_ca_bundle.setter
    def broker_ca_bundle(self, broker_ca_bundle):
        self._broker_ca_bundle = broker_ca_bundle

    @property
    def cert_file(self):
        """
        The file name of the client certificate in PEM format
        """
        return self._cert_file

    @cert_file.setter
    def cert_file(self, cert_file):
        self._cert_file = cert_file

    @property
    def private_key(self):
        """
        The file name of the client private key in PEM format
        """
        return self._private_key

    @private_key.setter
    def private_key(self, private_key):
        self._private_key = private_key

    @property
    def brokers(self):
        """
        A list of :class:`dxlclient.broker.Broker` objects representing brokers comprising the
        DXL fabric. When invoking the :func:`dxlclient.client.DxlClient.connect` method, the
        :class:`dxlclient.client.DxlClient` will attempt to connect to the closest broker.
        """
        return self._brokers

    @brokers.setter
    def brokers(self, brokers):
        self._brokers = brokers

    @property
    def incoming_message_queue_size(self):
        """
        The queue size for incoming messages (will block when queue is full)
        """
        return self._incoming_message_queue_size

    @incoming_message_queue_size.setter
    def incoming_message_queue_size(self, incoming_message_queue_size):
        self._incoming_message_queue_size = incoming_message_queue_size

    @property
    def incoming_message_queue_size(self):
        """
        The queue size for incoming messages (will block when queue is full).

        Defaults to ``1000``
        """
        return self._incoming_message_queue_size

    @incoming_message_queue_size.setter
    def incoming_message_queue_size(self, incoming_message_queue_size):
        self._incoming_message_queue_size = incoming_message_queue_size

    @property
    def incoming_message_thread_pool_size(self):
        """
        The thread pool size for incoming messages

        Defaults to ``1``
        """
        return self._incoming_message_thread_pool_size

    @incoming_message_thread_pool_size.setter
    def incoming_message_thread_pool_size(self, incoming_message_thread_pool_size):
        self._incoming_message_thread_pool_size = incoming_message_thread_pool_size

    @property
    def connect_retries(self):
        """
        The maximum number of connection attempts for each :class:`dxlclient.broker.Broker`
        specified in the :class:`dxlclient.client_config.DxlClientConfig`

        A value of ``-1`` indicates that the client will continue to retry without limit until it
        establishes a connection
        """
        return self._connect_retries

    @connect_retries.setter
    def connect_retries(self, connect_retries):
        """
        Sets the number of retries to perform when connecting. A value of -1
        indicates retry forever.

        :param connect_retries: The number of retries. A value of -1 indicates
                                retry forever.
        """
        self._connect_retries = connect_retries

    @property
    def keep_alive_interval(self):
        """
        The maximum period in seconds between communications with a connected :class:`dxlclient.broker.Broker`.
        If no other messages are being exchanged, this controls the rate at which the client will send ping
        messages to the :class:`dxlclient.broker.Broker`.

        Defaults to ``1800`` seconds (30 minutes)
        """
        return self._keep_alive_interval

    @property
    def reconnect_back_off_multiplier(self):
        """
        Multiples the current reconnect delay by this value on subsequent connect retries. For example, a current
        delay of 3 seconds with a multiplier of 2 would result in the next retry attempt being in 6 seconds.

        Defaults to ``2``
        """
        return self._reconnect_back_off_multiplier

    @reconnect_back_off_multiplier.setter
    def reconnect_back_off_multiplier(self, reconnect_back_off_multiplier):
        self._reconnect_back_off_multiplier = reconnect_back_off_multiplier

    @keep_alive_interval.setter
    def keep_alive_interval(self, keep_alive_interval):
        self._keep_alive_interval = keep_alive_interval

    @property
    def reconnect_delay(self):
        """
        The initial delay between retry attempts in seconds. The delay increases ("backs off")
        as subsequent connection attempts are made.

        Defaults to ``1`` second
        """
        return self._reconnect_delay

    @reconnect_delay.setter
    def reconnect_delay(self, reconnect_delay):
        self._reconnect_delay = reconnect_delay

    @property
    def reconnect_delay_max(self):
        """
        The maximum delay between connection retry attempts in seconds

        Defaults to ``60`` seconds (1 minute)
        """
        return self._reconnect_delay_max

    @reconnect_delay_max.setter
    def reconnect_delay_max(self, reconnect_delay_max):
        self._reconnect_delay_max = reconnect_delay_max

    @property
    def reconnect_delay_random(self):
        """
        Get the randomness delay percentage (between 0.0 and 1.0).
        The default value is 0.25
        """
        return self._reconnect_delay_random

    @reconnect_delay_random.setter
    def reconnect_delay_random(self, reconnect_delay_random):
        """
        Sets a randomness delay percentage (between 0.0 and 1.0). When
        calculating the reconnect delay, this percentage indicates how much
        randomness there should be in the current delay. For example, if the
        current delay is 100ms, a value of .25 would mean that the actual delay
        would be between 100ms and 125ms.

        :param reconnect_delay_random: The randomness delay percentage (between 0.0 and 1.0).
        """
        self._reconnect_delay_random = reconnect_delay_random

    @property
    def reconnect_when_disconnected(self):
        """
        Whether the client will continuously attempt to reconnect to the fabric if it becomes disconnected

        Defaults to ``True``
        """
        return self._reconnect_when_disconnected

    @reconnect_when_disconnected.setter
    def reconnect_when_disconnected(self, reconnect):
        self._reconnect_when_disconnected = reconnect

    def _set_brokers_from_json(self, broker_list):
        """
        Sets brokers list from JSON object.

        :param broker_list: Object containing dxl brokers. Should have this format:
        """
        brokers = _get_brokers(broker_list)
        if brokers is not None:
            self._brokers[:] = brokers

    def _get_sorted_broker_list_worker(self, broker):
        """Returns a sorted list of the brokers in this config."""
        broker._connect_to_broker()

    def _get_sorted_broker_list(self):
        """
        Returns the Broker list sorted by response time low to high.

        :returns: {@code list}: Sorted list of brokers.
        """
        threads = []

        for broker in self._brokers:
            # pylint: disable=invalid-name
            t = threading.Thread(target=self._get_sorted_broker_list_worker, args=[broker])
            threads.append(t)
            t.daemon = True
            t.start()

        for t in threads:
            t.join()

        return sorted(self._brokers, key=lambda b: (b._response_time is None, b._response_time))

    def _get_fastest_broker_worker(self, broker):
        """Calculate the fastest (smallest response time) broker."""
        broker._connect_to_broker()
        self._queue.put(broker)

    def _get_fastest_broker(self):
        """
        Returns the Broker with the lowest response time.

        :returns: {@code dxlclient.broker.Broker}: Fastest broker.
        """
        brokers = self._brokers
        self._queue = Queue()

        for broker in brokers:
            # pylint: disable=invalid-name
            t = threading.Thread(target=self._get_fastest_broker_worker, args=[broker])
            t.daemon = True
            t.start()

        return self._queue.get(timeout=15)

    @staticmethod
    def create_dxl_config_from_file(dxl_config_file):
        """

        This method allows creation of a :class:`DxlClientConfig` object from a
        specified configuration file. The information contained in the file has a one-to-one
        correspondence with the :class:`DxlClientConfig` constructor.

        .. code-block:: python

            [Certs]
            BrokerCertChain=c:\\\\certs\\\\brokercerts.crt
            CertFile=c:\\\\certs\\\\client.crt
            PrivateKey=c:\\\\certs\\\\client.key

            [Brokers]
            mybroker=mybroker;8883;mybroker.mcafee.com;192.168.1.12
            mybroker2=mybroker2;8883;mybroker2.mcafee.com;192.168.1.13

        The configuration file can be loaded as follows:

        .. code-block:: python

            from dxlclient.client_config import DxlClientConfig

            config = DxlClientConfig.create_dxl_config_from_file("c:\\\\certs\\\\dxlclient.cfg")

        :param dxl_config_file: Path to the configuration file
        :return: A :class:`DxlClientConfig` object corresponding to the specified configuration file
        """
        config_parser = _DxlConfigParser()

        if not config_parser.read(dxl_config_file):
            raise Exception("Can't parse config file")

        config_file_path = path.dirname(dxl_config_file)
        cert_file = DxlClientConfig._get_file_path(config_file_path, config_parser.get("Certs", "CertFile"))
        private_key = DxlClientConfig._get_file_path(config_file_path, config_parser.get("Certs", "PrivateKey"))
        cert_chain = DxlClientConfig._get_file_path(config_file_path, config_parser.get("Certs", "BrokerCertChain"))
        client_id = config_parser.get("General", "ClientId", False)

        client_config = DxlClientConfig(broker_ca_bundle=cert_chain,
                                        cert_file=cert_file, private_key=private_key, brokers=[])
        if client_id:
            client_config._client_id = client_id

        broker_list = {}
        try:
            brokers = config_parser.items("Brokers")
            for broker in brokers:
                broker_list[broker[0]] = broker[1]
        except NoSectionError:
            logger.warning("Brokers not defined in config file")

        if len(broker_list) is 0:
            logger.warning("Broker list is empty")

        client_config._set_brokers_from_json(broker_list)
        return client_config

    @staticmethod
    def _get_file_path(config_path, cert_file_path):
        if not path.isfile(cert_file_path) and not path.isabs(cert_file_path):
            file_path = path.join(config_path, cert_file_path)
            if path.isfile(file_path):
                cert_file_path = file_path
        return cert_file_path



# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2017 McAfee Inc. - All Rights Reserved.
################################################################################

"""
The messages module contains the different types of messages that are transmitted
over the Data Exchange Layer (DXL) fabric.

- :class:`Event`
- :class:`Request`
- :class:`Response`
- :class:`ErrorResponse`

:class:`Event` messages are sent using the :func:`dxlclient.client.DxlClient.send_event` method of a client instance.
Event messages are sent by one publisher and received by one or more recipients that are currently
subscribed to the :attr:`Message.destination_topic` associated with the event (otherwise known as one-to-many).

:class:`Request` messages are sent using the :func:`dxlclient.client.DxlClient.sync_request` and
:func:`dxlclient.client.DxlClient.async_request` methods of a client instance. Request messages are used when
invoking a method on a remote service. This communication is one-to-one where a client sends a request to a
service instance and in turn receives a response.

:class:`Response` messages are sent by service instances upon receiving :class:`Request` messages. Response messages
are sent using the :func:`dxlclient.client.DxlClient.send_response` method of a client instance. Clients that are
invoking the service (sending a request) will receive the response as a return value of the
:func:`dxlclient.client.DxlClient.sync_request` method of a client instance or via the
:class:`dxlclient.callbacks.ResponseCallback` callback when invoking the asynchronous method,
:func:`dxlclient.client.DxlClient.async_request`.

:class:`ErrorResponse` messages are sent by the DXL fabric itself or service instances upon receiving :class:`Request`
messages. The error response may indicate the inability to locate a service to handle the request or an internal
error within the service itself. Error response messages are sent using the
:func:`dxlclient.client.DxlClient.send_response` method of a client instance.

**NOTE:** Some services may chose to not send a :class:`Response` message when receiving a :class:`Request`.
This typically occurs if the service is being used to simply collect information from remote clients. In this scenario,
the client should use the asynchronous form for sending requests,
:func:`dxlclient.client.DxlClient.async_request`
"""

import msgpack  # pylint: disable=import-error
from io import BytesIO
from abc import ABCMeta, abstractproperty, abstractmethod

from dxlclient import _BaseObject
from dxlclient._uuid_generator import UuidGenerator
from dxlclient.exceptions import DxlException

# pylint: disable=too-many-instance-attributes
class Message(_BaseObject):
    """
    The base class for the different Data Exchange Layer (DXL) message types
    """
    __metaclass__ = ABCMeta

    # The message version
    MESSAGE_VERSION = 0

    MESSAGE_TYPE_REQUEST = 0
    """The numeric type identifier for the :class:`Request` message type"""
    MESSAGE_TYPE_RESPONSE = 1
    """The numeric type identifier for the :class:`Response` message type"""
    MESSAGE_TYPE_EVENT = 2
    """The numeric type identifier for the :class:`Event` message type"""
    MESSAGE_TYPE_ERROR = 3
    """The numeric type identifier for the :class:`ErrorResponse` message type"""

    def __init__(self, destination_topic):
        """
        Constructor parameters:

        :param destination_topic: The topic to publish the message to
        """
        super(Message, self).__init__()

        # The version of the message
        self._version = self.MESSAGE_VERSION
        # The unique identifier for the message
        self._message_id = UuidGenerator.generate_id_as_string()
        # The identifier for the client that is the source of the message
        self._source_client_id = ""
        # The GUID for the broker that is the source of the message
        self._source_broker_id = ""
        # The channel that the message is published on
        self._destination_topic = destination_topic
        # The payload to send with the message
        self._payload = bytes()
        # The set of broker GUIDs to deliver the message to
        self._broker_ids = []
        # The set of client GUIDs to deliver the message to
        self._client_ids = []

    def __del__(self):
        """destructor"""
        super(Message, self).__del__()

    @property
    def version(self):
        """
        The version of the DXL message (used to determine the features that are available)
        """
        return self._version

    @property
    def message_id(self):
        """
        Unique identifier for the message (UUID)
        """
        return self._message_id

    @property
    def destination_topic(self):
        """
        The topic to publish the message to
        """
        return self._destination_topic

    @destination_topic.setter
    def destination_topic(self, topic):
        self._destination_topic = topic

    @property
    def payload(self):
        """
        The application-specific payload of the message (bytes)
        """
        return self._payload

    @payload.setter
    def payload(self, payload):
        self._payload = payload

    @property
    def source_client_id(self):
        """
        The identifier of the DXL client that sent the message (set by the broker that initially receives the message)
        """
        return self._source_client_id

    @property
    def source_broker_id(self):
        """
        The identifier of the DXL broker that the message's originating client is connected to
        (set by the initial broker)
        """
        return self._source_broker_id

    @abstractproperty
    def message_type(self):
        """
        The numeric type of the message
        """
        return None

    @property
    def broker_ids(self):
        """
        The set of broker identifiers that the message is to be routed to. Setting this value will limit
        which brokers the message will be delivered to. This can be used in conjunction with :func:`client_ids`.
        """
        return self._broker_ids

    @broker_ids.setter
    def broker_ids(self, broker_guids=None):
        if broker_guids is None:
            broker_guids = []
        self._broker_ids = broker_guids

    @property
    def client_ids(self):
        """
        The set of client identifiers that the message is to be routed to. Setting this value will limit
        which clients the message will be delivered to. This can be used in conjunction with :func:`broker_ids`.
        """
        return self._client_ids

    @client_ids.setter
    def client_ids(self, client_guids=None):
        if client_guids is None:
            client_guids = []
        self._client_ids = client_guids

    @abstractmethod
    def _pack_message(self, packer, buf):
        """
        Converts the message to an array of bytes and writes them to buf.

        :param packer: Packer object.
        :param buf: Object to which the bytes will be written. Must have a write method.
        """
        buf.write(packer.pack(self._message_id))
        buf.write(packer.pack(self._source_client_id))
        buf.write(packer.pack(self._source_broker_id))
        buf.write(packer.pack(self._broker_ids))
        buf.write(packer.pack(self._client_ids))
        buf.write(packer.pack(self._payload))

    @abstractmethod
    def _unpack_message(self, unpacker):
        """
        Creates a concrete message from `unpacker`.

        :param unpacker: Unpacker object.
        """
        self._message_id = unpacker.next()
        self._source_client_id = unpacker.next()
        self._source_broker_id = unpacker.next()
        self._broker_ids = unpacker.next()
        self._client_ids = unpacker.next()
        self._payload = unpacker.next()

    def _to_bytes(self):
        """
        Converts the message to an array of bytes and returns it.

        :returns: {@code BytesIO} object.
        """
        buf = BytesIO()
        packer = msgpack.Packer()
        buf.write(packer.pack(self.version))
        buf.write(packer.pack(self.message_type))
        self._pack_message(packer, buf)
        return buf.getvalue()

    @staticmethod
    def _from_bytes(raw):
        """
        Converts the specified array of bytes to a concrete message instance
        (request, response, error, etc.) and returns it.

        :param raw: {@code list} of bytes.
        :returns: {@link dxlclient.message.Message} object.
        """
        buf = BytesIO(raw)
        buf.seek(0)
        unpacker = msgpack.Unpacker(buf)
        version = unpacker.next()
        message_type = unpacker.next()

        message = None
        if message_type == Message.MESSAGE_TYPE_REQUEST:
            message = Request(destination_topic="")
        elif message_type == Message.MESSAGE_TYPE_ERROR:
            message = ErrorResponse(request=None)
        elif message_type == Message.MESSAGE_TYPE_RESPONSE:
            message = Response(request="")
        elif message_type == Message.MESSAGE_TYPE_EVENT:
            message = Event(destination_topic="")

        if message is not None:
            message._version = version
            message._unpack_message(unpacker)
            return message

        raise DxlException("Unknown message type: " + message_type)


class Request(Message):
    """
    :class:`Request` messages are sent using the :func:`dxlclient.client.DxlClient.sync_request` and
    :func:`dxlclient.client.DxlClient.async_request` methods of a client instance. Request messages are used when
    invoking a method on a remote service. This communication is one-to-one where a client sends a request to a
    service instance and in turn receives a response.
    """

    def __init__(self, destination_topic):
        """
        Constructor parameters:

        :param destination_topic: The topic to publish the request to
        """
        super(Request, self).__init__(destination_topic)
        # The topic used to reply to this request
        self._reply_to_topic = None
        # The service id
        self._service_id = ""

    @property
    def message_type(self):
        """
        The numeric type of the message
        """
        return Message.MESSAGE_TYPE_REQUEST

    @property
    def reply_to_topic(self):
        """
        The topic that the :class:`Response` to this :class:`Request` will be sent to
        """
        return self._reply_to_topic

    @reply_to_topic.setter
    def reply_to_topic(self, topic):
        self._reply_to_topic = topic

    @property
    def service_id(self):
        """
        The identifier of the service that this request will be routed to. If an identifier is not specified,
        the initial broker that receives the request will select the service to handle the request (round-robin by
        default).
        """
        return self._service_id

    @service_id.setter
    def service_id(self, service_id):
        self._service_id = service_id

    def _pack_message(self, packer, buf):
        """
        Converts the message to an array of bytes and writes them to buf.

        :param packer: Packer object.
        :param buf: Object to which the bytes will be written. Must have a write method.
        """
        super(Request, self)._pack_message(packer, buf)
        buf.write(packer.pack(self._reply_to_topic))
        buf.write(packer.pack(self._service_id))

    def _unpack_message(self, unpacker):
        """
        Creates a concrete message from `unpacker`.

        :param unpacker: Unpacker object.
        """
        super(Request, self)._unpack_message(unpacker)
        self._reply_to_topic = unpacker.next()
        self._service_id = unpacker.next()


class Response(Message):
    """
    :class:`Response` messages are sent by service instances upon receiving :class:`Request` messages. Response messages
    are sent using the :func:`dxlclient.client.DxlClient.send_response` method of a client instance. Clients that are
    invoking the service (sending a request) will receive the response as a return value of the
    :func:`dxlclient.client.DxlClient.sync_request` method of a client instance or via the
    :class:`dxlclient.callbacks.ResponseCallback` callback when invoking the asynchronous method,
    :func:`dxlclient.client.DxlClient.async_request`.
    """

    def __init__(self, request):
        """
        Constructor parameters:

        :param request: The :class:`Request` message that this is a response for
        """
        if isinstance(request, Request):
            super(Response, self).__init__(request.reply_to_topic)
            # The request (only available when sending the response)
            self._request = request
            # The identifier for the request message that this is a response for
            self._request_message_id = request.message_id

            self._service_id = request.service_id
            id = request.source_client_id
            if id is not None and len(id) > 0:
                self.client_ids = [id]
            id = request.source_broker_id
            if id is not None and len(id) > 0:
                self.broker_ids = [id]
        else:
            super(Response, self).__init__(destination_topic="")
            # The request (only available when sending the response)
            self._request = None
            # The identifier for the request message that this is a response for
            self._request_message_id = None
            # The identifier of the service that processed the request
            self._service_id = ""

    @property
    def message_type(self):
        """
        The numeric type of the message
        """
        return Message.MESSAGE_TYPE_RESPONSE

    @property
    def request_message_id(self):
        """
        Unique identifier (UUID) for the :class:`Request` message that this message is a
        response for. This is used by the invoking :class:`dxlclient.client.DxlClient` to
        correlate an incoming :class:`Response` message with the :class:`Request` message that
        was initially sent by the client.
        """
        return self._request_message_id

    @property
    def request(self):
        """
        The :class:`Request` message that this is a response for
        """
        return self._request

    @property
    def service_id(self):
        """
        The identifier of the service that sent this response (the service that the corresponding
        :class:`Request` was routed to).
        """
        return self._service_id

    def _pack_message(self, packer, buf):
        """
        Converts the message to an array of bytes and writes them to buf.

        :param packer: Packer object.
        :param buf: Object to which the bytes will be written. Must have a write method.
        """
        super(Response, self)._pack_message(packer, buf)
        buf.write(packer.pack(self._request_message_id))
        buf.write(packer.pack(self._service_id))

    def _unpack_message(self, unpacker):
        """
        Creates a concrete message from `unpacker`.

        :param unpacker: Unpacker object.
        """
        super(Response, self)._unpack_message(unpacker)
        self._request_message_id = unpacker.next()
        self._service_id = unpacker.next()


class Event(Message):
    """
    :class:`Event` messages are sent using the :func:`dxlclient.client.DxlClient.send_event` method of a client instance.
    Event messages are sent by one publisher and received by one or more recipients that are currently
    subscribed to the :attr:`Message.destination_topic` associated with the event (otherwise known as one-to-many).
    """
    def __init__(self, destination_topic):
        """
        Constructor parameters:

        :param destination_topic: The topic to publish the event to
        """
        super(Event, self).__init__(destination_topic)

    @property
    def message_type(self):
        """
        The numeric type of the message
        """
        return Message.MESSAGE_TYPE_EVENT

    def _pack_message(self, packer, buf):
        """
        Converts the message to an array of bytes and writes them to buf.

        :param packer: Packer object.
        :param buf: Object to which the bytes will be written. Must have a write method.
        """
        super(Event, self)._pack_message(packer, buf)

    def _unpack_message(self, unpacker):
        """
        Creates a concrete message from `unpacker`.

        :param unpacker: Unpacker object.
        """
        super(Event, self)._unpack_message(unpacker)


class ErrorResponse(Response):
    """
    :class:`ErrorResponse` messages are sent by the DXL fabric itself or service instances upon receiving
    :class:`Request` messages. The error response may indicate the inability to locate a service to handle the
    request or an internal error within the service itself. Error response messages are sent using the
    :func:`dxlclient.client.DxlClient.send_response` method of a client instance.
    """

    def __init__(self, request, error_code=0, error_message=""):
        """
        Constructor parameters:

        :param request: The :class:`Request` message that this is a response for
        :param error_code: The numeric error code
        :param error_message: The textual error message
        """
        super(ErrorResponse, self).__init__(request)
        # The error code
        self._error_code = error_code
        # The error message
        self._error_message = error_message

    @property
    def error_code(self):
        """
        The numeric error code for the error response
        """
        return self._error_code

    @property
    def error_message(self):
        """
        The textual error message for the error response
        """
        return self._error_message

    @property
    def message_type(self):
        """
        The numeric type of the message
        """
        return Message.MESSAGE_TYPE_ERROR

    def _pack_message(self, packer, buf):
        """
        Converts the message to an array of bytes and writes them to buf.

        :param packer: Packer object.
        :param buf: Object to which the bytes will be written. Must have a write method.
        """
        super(ErrorResponse, self)._pack_message(packer, buf)
        buf.write(packer.pack(self._error_code))
        buf.write(packer.pack(self._error_message))

    def _unpack_message(self, unpacker):
        """
        Creates a concrete message from `unpacker`.

        :param unpacker: Unpacker object.
        """
        super(ErrorResponse, self)._unpack_message(unpacker)
        self._error_code = unpacker.next()
        self._error_message = unpacker.next()



# This sample demonstrates wrapping an existing service and exposing it on the
# DXL fabric.
#
# In this particular case, the "openweathermap.org" current weather data is
# exposed as a DXL service. This service wrapper delegates to the
# OpenWeatherMap REST API.
#
# openweather_common.py must be edited to include the OpenWeatherMap API
# key (see http://openweathermap.org/appid)

import logging
import os
import sys
import time
import urllib2

from dxlclient.callbacks import RequestCallback
from dxlclient.client import DxlClient
from dxlclient.client_config import DxlClientConfig
from dxlclient.message import ErrorResponse, Response
from dxlclient.service import ServiceRegistrationInfo

from openweather_common import *

# Import common logging and configuration
sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/..")
from common import *

# Configure local logger
logging.getLogger().setLevel(logging.INFO)
logger = logging.getLogger(__name__)

# The OpenWeatherMap "Current Weather" URL (see http://openweathermap.org/current)
CURRENT_WEATHER_URL = "http://api.openweathermap.org/data/2.5/weather?{0}&APPID={1}"

# Create DXL configuration from file
config = DxlClientConfig.create_dxl_config_from_file(CONFIG_FILE)

# Create the client
with DxlClient(config) as client:

    # Connect to the fabric
    client.connect()

    #
    # Register the service
    #

    # Create "Current Weather" incoming request callback
    class CurrentWeatherCallback(RequestCallback):
        def on_request(self, request):
            try:
                # Extract information from request
                query = request.payload.decode(encoding="UTF-8")
                logger.info("Service received request payload: " + query)

                # Send HTTP request to OpenWeatherMap
                req = urllib2.Request(
                    CURRENT_WEATHER_URL.format(query, API_KEY), None,
                    {'Content-Type': 'text/json'})
                f = urllib2.urlopen(req)
                weather_response = f.read()
                f.close()

                # Create the response message
                response = Response(request)
                # Populate the response payload
                response.payload = weather_response.encode(encoding="UTF-8")
                # Send the response
                client.send_response(response)

            except Exception as ex:
                print str(ex)
                # Send error response
                client.send_response(ErrorResponse(request, error_message=str(ex).encode(encoding="UTF-8")))

    # Create service registration object
    info = ServiceRegistrationInfo(client, SERVICE_NAME)

    # Add a topic for the service to respond to
    info.add_topic(SERVICE_CURRENT_WEATHER_TOPIC, CurrentWeatherCallback())

    # Register the service with the fabric (wait up to 10 seconds for registration to complete)
    client.register_service_sync(info, 10)

    logger.info("Weather service is running...")

    # Wait forever
    while True:
        time.sleep(60)



# This sample demonstrates how to register a callback to receive Event messages
# from the DXL fabric. Once the callback is registered, the sample sends a
# set number of Event messages to the fabric and waits for them all to be
# received by the callback.

import logging
import os
import sys
import time
from threading import Condition

from dxlclient.callbacks import EventCallback
from dxlclient.client import DxlClient
from dxlclient.client_config import DxlClientConfig
from dxlclient.message import Event

# Import common logging and configuration
sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/..")
from common import *

# Configure local logger
logging.getLogger().setLevel(logging.ERROR)
logger = logging.getLogger(__name__)

# The topic to publish to
EVENT_TOPIC = "/isecg/sample/basicevent"

# The total number of events to send
TOTAL_EVENTS = 1000

# Condition/lock used to protect changes to counter
event_count_condition = Condition()

# The events received (use an array so we can modify in callback)
event_count = [0]

# Create DXL configuration from file
config = DxlClientConfig.create_dxl_config_from_file(CONFIG_FILE)

# Create the client
with DxlClient(config) as client:

    # Connect to the fabric
    client.connect()

    #
    # Register callback and subscribe
    #

    # Create and add event listener
    class MyEventCallback(EventCallback):
        def on_event(self, event):
            with event_count_condition:
                # Print the payload for the received event
                print "Received event: " + event.payload.decode()
                # Increment the count
                event_count[0] += 1
                # Notify that the count was increment
                event_count_condition.notify_all()

    # Register the callback with the client
    client.add_event_callback(EVENT_TOPIC, MyEventCallback())

    #
    # Send events
    #

    # Record the start time
    start = time.time()

    # Loop and send the events
    for event_id in range(TOTAL_EVENTS):
        # Create the event
        event = Event(EVENT_TOPIC)
        # Set the payload
        event.payload = str(event_id).encode()
        # Send the event
        client.send_event(event)

    # Wait until all events have been received
    print "Waiting for events to be received..."
    with event_count_condition:
        while event_count[0] < TOTAL_EVENTS:
            event_count_condition.wait()

    # Print the elapsed time
    print "Elapsed time (ms): " + str((time.time() - start) * 1000)


from threading import Condition
from dxlclient.test.test_service import TestService
from dxlclient import UuidGenerator, Request, ResponseCallback, ServiceRegistrationInfo
from dxlclient.test.base_test import BaseClientTest, atomize
from nose.plugins.attrib import attr


class AsyncRequestTests(BaseClientTest):

    # The number of requests to send
    REQ_COUNT = 100
    MAX_RESPONSE_WAIT = 1 * 60

    response_count = 0
    outstanding_requests = []
    # Conditions
    response_condition = Condition()
    outstanding_requests_condition = Condition()

    @atomize(outstanding_requests_condition)
    def append_outstanding_request(self, message_id):
        self.outstanding_requests.append(message_id)

    @atomize(outstanding_requests_condition)
    def remove_outstanding_request(self, message_id):
        self.outstanding_requests.remove(message_id)

    #
    # Tests the asynchronous request methods of the DxlClient.
    #
    @attr('system')
    def test_execute_async_request(self):

        # Create and register a response callback. Once the request has been processed
        # by the service, the requesting client will receive a notification via this
        # callback. At that point, we note that the request has been responded to (remove
        # it from the set of outstanding requests), and increment the count of responses
        # we have received via the callbacks.
        def my_response(response):
            with self.response_condition:
                # Increment count of responses received
                self.response_count += 1
                # Remove from outstanding requests
                self.remove_outstanding_request(response.request_message_id)
                # Notify that a response has been received (are we done yet?)
                self.response_condition.notify_all()
        rc = ResponseCallback()
        rc.on_response = my_response

        with self.create_client(max_retries=0) as client:
            try:
                client.connect()
                # Create a test service that responds to requests on a particular topic.
                test_service = TestService(client, 1)
                topic = UuidGenerator.generate_id_as_string()
                reg_info = ServiceRegistrationInfo(client, "async_request_runner_service")
                reg_info.add_topic(topic, test_service)
                # Register the service
                client.register_service_sync(reg_info, self.DEFAULT_TIMEOUT)
                # Add a response callback (not channel-specific)
                client.add_response_callback("", rc)

                for i in range(0, self.REQ_COUNT):
                    # Send a request without specifying a callback for the current request
                    req = Request(topic)
                    self.append_outstanding_request(req.message_id)
                    client.async_request(req)

                    # Send a request with a specific callback that is to receive the response.
                    # The response will be received by two callbacks.
                    req = Request(topic)
                    self.append_outstanding_request(req.message_id)
                    client.async_request(req, response_callback=rc)

                # Wait until all the responses are received via the response callbacks.
                # The "times 3" is due to the fact that 20000 requests were sent in total.
                # 20000 were handled via the global callback, an additional 10000 were handled
                # via the callback explicitly passed in the second set of requests.
                with self.response_condition:
                    while self.response_count != self.REQ_COUNT * 3:
                        current_count = self.response_count
                        self.response_condition.wait(self.MAX_RESPONSE_WAIT)
                        if current_count == self.response_count:
                            self.fail("Response wait timeout")

                # Make sure there are no outstanding requests
                self.assertEqual(0, len(self.outstanding_requests))
                print "Async request test: PASSED"

            except Exception, ex:
                print ex.message
                raise ex



from functools import wraps
from unittest import TestCase
from dxlclient import DxlClientConfig, DxlClient
import os


def atomize(lock):
    def decorator(wrapped):
        @wraps(wrapped)
        def _wrapper(*args, **kwargs):
            with lock:
                return wrapped(*args, **kwargs)
        return _wrapper
    return decorator


class BaseClientTest(TestCase):
    DEFAULT_TIMEOUT = 5 * 60
    DEFAULT_RETRIES = 3
    POST_OP_DELAY = 8
    REG_DELAY = 60

    def create_client(self, max_retries=DEFAULT_RETRIES, incoming_message_thread_pool_size = 1):
        config = DxlClientConfig.create_dxl_config_from_file(os.path.dirname(os.path.abspath(__file__)) +
                                                              "/client_config.cfg")
        config.incoming_message_thread_pool_size = incoming_message_thread_pool_size

        config.connect_retries = max_retries
        return DxlClient(config)



import logging
import os
import sys

from dxlclient.callbacks import EventCallback
from dxlclient.client import DxlClient
from dxlclient.client_config import DxlClientConfig

# Import common logging and configuration
sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/..")
from common import *

# Configure local logger
logger = logging.getLogger(__name__)

# Sample topic to fire Events on
EVENT_TOPIC = "/isecg/sample/event"

# Sample Event Subscriber
try:
    # Create DxlClientConfig from expected sample configuration file
    logger.info("Event Subscriber - Load DXL config from: %s", CONFIG_FILE)
    config = DxlClientConfig.create_dxl_config_from_file(CONFIG_FILE)

    # Initialize DXL client using our configuration
    logger.info("Event Subscriber - Creating DXL Client")
    with DxlClient(config) as client:

        # Connect to DXL Broker
        logger.info("Event Subscriber - Connecting to Broker")
        client.connect()

        # Event callback class to handle incoming DXL Events
        class MyEventCallback(EventCallback):
            def on_event(self, event):
                # Extract information from Event payload, in this sample we expect it is UTF-8 encoded
                logger.info("Event Subscriber - Event received:\n   Topic: %s\n   Payload: %s",
                            event.destination_topic, event.payload.decode())

        # Add Event callback to DXL client
        logger.info("Adding Event callback function to Topic: %s", EVENT_TOPIC)
        client.add_event_callback(EVENT_TOPIC, MyEventCallback())

        # Wait for DXL Events
        while True:
            print "   Enter 9 to quit"
            input = raw_input("   Enter value: ")

            try:
                option = int(input)
            except:
                option = input

            # Option: Exit the loop
            if option == 9:
                break

            # Invalid input
            else:
                logger.info("Event Subscriber - Invalid input: %s", option)

except Exception as e:
    logger.info("Event Subscriber - Exception: %s", e.message)
    exit(1)


from dxlclient import Request, UuidGenerator
from dxlclient.test.base_test import BaseClientTest
from nose.plugins.attrib import attr

@attr('system')
class SubsCountTest(BaseClientTest):

    def _create_request(self, topic):
        req = Request("/mcafee/service/dxl/broker/subs")
        req.payload = "{\"topic\":\"" + topic + "\"}"
        return req

    @attr('system')
    def test_subs_count(self):
        clients = []

        with self.create_client() as client:
            client.connect()

            for i in range(6):
                c = self.create_client()
                c.connect()
                clients.append(c)

            random1 = UuidGenerator.generate_id_as_string()
            random2 = UuidGenerator.generate_id_as_string()
            topic1 = "/foo/bar/" + random1 + "/" + random2
            clients[0].subscribe(topic1)
            clients[1].subscribe(topic1)
            clients[2].subscribe(topic1)
            clients[3].subscribe("/foo/bar/" + random1 + "/#")
            clients[4].subscribe("/foo/+/" + random1 + "/#")

            topic2 = "/foo/baz/" + random2
            clients[1].subscribe(topic2)
            clients[2].subscribe(topic2)
            clients[5].subscribe("#")

            # Topic 1
            req = self._create_request(topic1)
            resp = client.sync_request(req, 5)
            pl = str(resp.payload)
            self.assertTrue(":6}" in pl)

            # Topic 2
            req = self._create_request(topic2)
            resp = client.sync_request(req, 5)
            pl = str(resp.payload)
            self.assertTrue(":3}" in pl)

            print "Test passed!"



# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2014 McAfee Inc. - All Rights Reserved.
################################################################################

# Run with python -m unittest dxlclient.test.test_callback_manager

import unittest

import dxlclient.callbacks as callbacks
import dxlclient._callback_manager as callback_manager


class MockRequestCallback(callbacks.RequestCallback):
    def on_request(self, request):
        pass


class MockResponseCallback(callbacks.ResponseCallback):
    def on_response(self, response):
        pass


class MockEventCallback(callbacks.EventCallback):
    def on_event(self, event):
        pass


class CallbackManagerTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_request_callback_manager_with_valid_callback(self):
        cbm = callback_manager._RequestCallbackManager()
        cbm.add_callback("/test", MockRequestCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("/test")))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.add_callback(callback=MockRequestCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("")))
        self.assertEquals(2, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", MockRequestCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback(callback=MockRequestCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get(""))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_request_callback_manager_with_invalid_callback(self):
        cbm = callback_manager._RequestCallbackManager()
        with self.assertRaises(ValueError):
            cbm.add_callback("/test", MockResponseCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_request_callback_manager_with_double_registration(self):
        cbm = callback_manager._RequestCallbackManager()
        cbm.add_callback("/test", MockRequestCallback)
        cbm.add_callback("/test", MockRequestCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", MockRequestCallback)
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_request_callback_manager_with_valid_callback_instance(self):
        cbm = callback_manager._RequestCallbackManager()
        cb = MockRequestCallback()
        cbm.add_callback("/test", cb)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("/test")))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.add_callback(callback=cb)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("")))
        self.assertEquals(2, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback(callback=cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get(""))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_request_callback_manager_with_invalid_callback_instance(self):
        cbm = callback_manager._RequestCallbackManager()
        cb = MockResponseCallback()
        with self.assertRaises(ValueError):
            cbm.add_callback("/test", cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_response_callback_manager_with_valid_callback(self):
        cbm = callback_manager._ResponseCallbackManager()
        cbm.add_callback("/test", MockResponseCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("/test")))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.add_callback(callback=MockResponseCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("")))
        self.assertEquals(2, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", MockResponseCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback(callback=MockResponseCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get(""))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_response_callback_manager_with_invalid_callback(self):
        cbm = callback_manager._ResponseCallbackManager()
        with self.assertRaises(ValueError):
            cbm.add_callback("/test", MockEventCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_response_callback_manager_with_double_registration(self):
        cbm = callback_manager._ResponseCallbackManager()
        cbm.add_callback("/test", MockResponseCallback)
        cbm.add_callback("/test", MockResponseCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", MockResponseCallback)
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_response_callback_manager_with_valid_callback_instance(self):
        cbm = callback_manager._ResponseCallbackManager()
        cb = MockResponseCallback()
        cbm.add_callback("/test", cb)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("/test")))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.add_callback(callback=cb)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("")))
        self.assertEquals(2, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback(callback=cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get(""))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_response_callback_manager_with_invalid_callback_instance(self):
        cbm = callback_manager._ResponseCallbackManager()
        cb = MockEventCallback()
        with self.assertRaises(ValueError):
            cbm.add_callback("/test", cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_event_callback_manager_with_valid_callback(self):
        cbm = callback_manager._EventCallbackManager()
        cbm.add_callback("/test", MockEventCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("/test")))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.add_callback(callback=MockEventCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("")))
        self.assertEquals(2, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", MockEventCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback(callback=MockEventCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get(""))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_event_callback_manager_with_invalid_callback(self):
        cbm = callback_manager._EventCallbackManager()
        with self.assertRaises(ValueError):
            cbm.add_callback("/test", MockRequestCallback)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_event_callback_manager_with_double_registration(self):
        cbm = callback_manager._EventCallbackManager()
        cbm.add_callback("/test", MockEventCallback)
        cbm.add_callback("/test", MockEventCallback)
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", MockEventCallback)
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_event_callback_manager_with_valid_callback_instance(self):
        cbm = callback_manager._EventCallbackManager()
        cb = MockEventCallback()
        cbm.add_callback("/test", cb)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("/test")))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.add_callback(callback=cb)
        self.assertEquals(1, len(cbm.callbacks_by_channel.get("")))
        self.assertEquals(2, len(cbm.callbacks_by_channel))
        cbm.remove_callback("/test", cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(1, len(cbm.callbacks_by_channel))
        cbm.remove_callback(callback=cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get(""))
        self.assertEquals(0, len(cbm.callbacks_by_channel))

    def test_event_callback_manager_with_invalid_callback_instance(self):
        cbm = callback_manager._EventCallbackManager()
        cb = MockRequestCallback()
        with self.assertRaises(ValueError):
            cbm.add_callback("/test", cb)
        self.assertEquals(None, cbm.callbacks_by_channel.get("/test"))
        self.assertEquals(0, len(cbm.callbacks_by_channel))



from threading import Condition
from dxlclient.test.base_test import BaseClientTest
from dxlclient.test.test_service import TestService
from nose.plugins.attrib import attr
from dxlclient import ServiceRegistrationInfo, Request, ErrorResponse
from thread_executor import ThreadRunExecutor

@attr('system')
class SyncRequestTests(BaseClientTest):

    # The number of requests to send
    REQUEST_COUNT = 500
    # Maximum time to wait for the test to complete
    MAX_WAIT = 5 * 60
    # Maximum time to wait for a response
    RESPONSE_WAIT = 60

    response_count = 0
    response_count_condition = Condition()

    event_condition = Condition()

    #
    # Tests the synchronous request methods of the DxlClient.
    #
    @attr('system')
    def test_execute_sync_request(self):
        with self.create_client(max_retries=0) as client:
            # Create a test service that responds to requests on a particular topic.
            test_service = TestService(client, 1)
            client.connect()
            topic = "event_testing"  # UuidGenerator.generate_id_as_string()
            reg_info = ServiceRegistrationInfo(client, "sync_request_runner_service")
            reg_info.add_topic(topic, test_service)
            # Register the service
            client.register_service_sync(reg_info, self.DEFAULT_TIMEOUT)

            executor = ThreadRunExecutor(self.REQUEST_COUNT)

            # Sends synchronous requests with a unique thread for each request. Ensure that the
            # response that is received corresponds with the request that was sent. Also, keep
            # track of the total number of responses received.
            def run():
                try:
                    request = Request(topic)
                    response = client.sync_request(request, timeout=self.RESPONSE_WAIT)
                    self.assertNotIsInstance(response, ErrorResponse)
                    self.assertEquals(request.message_id, response.request_message_id)

                    with self.response_count_condition:
                        self.response_count += 1
                        if self.response_count % 100 == 0:
                            print self.response_count
                        self.response_count_condition.notify_all()
                except Exception, e:
                    print e.message
                    raise e

            executor.execute(run)

            # Wait for all of the requests to complete
            with self.response_count_condition:
                while self.response_count != self.REQUEST_COUNT:
                    current_count = self.response_count
                    self.response_count_condition.wait(self.MAX_WAIT)
                    if current_count == self.response_count:
                        self.fail("Request wait timeout.")

            self.assertEquals(self.REQUEST_COUNT, self.response_count)
